import re
import logging
import json

from typing import List, Tuple, Set, Dict
from dataclasses import dataclass

from prefixcommons import curie_util

from ontobio.io import assocparser
from ontobio.io import parser_version_regex
from ontobio.io.assocparser import ENTITY, EXTENSION, ANNOTATION, Report
from ontobio.io import qc
from ontobio.io import entityparser
from ontobio.io import entitywriter
from ontobio.model import association
from ontobio.model import collections
from ontobio.ecomap import EcoMap
from ontobio.rdfgen import relations
from ontobio.ontol import Ontology

import dateutil.parser
import functools

import click

logger = logging.getLogger(__name__)




gaf_line_validators = {
    "default": assocparser.ColumnValidator(),
    "qualifier2_1": assocparser.Qualifier2_1(),
    "qualifier2_2": assocparser.Qualifier2_2(),
    "curie": assocparser.CurieValidator(),
    "taxon": assocparser.TaxonValidator()
}


def protein_complex_sublcass_closure(ontology: Ontology) -> Set[str]:
    protein_containing_complex = association.Curie(namespace="GO", identity="0032991")
    children_of_complexes = set(ontology.descendants(str(protein_containing_complex), relations=["subClassOf"], reflexive=True))
    return children_of_complexes


class GafParser(assocparser.AssocParser):
    """
    Parser for GO GAF format
    """

    ANNOTATION_CLASS_COLUMN=4

    def __init__(self, config=None, group="unknown", dataset="unknown", bio_entities=None):
        """
        Arguments:
        ---------

        config : a AssocParserConfig object
        """
        self.config = config
        self.group = group
        self.version = None
        self.default_version = "2.1"
        self.bio_entities = bio_entities
        if self.bio_entities is None:
            self.bio_entities = collections.BioEntities(dict())

        self.cell_component_descendants_closure = None

        if config is None:
            self.config = assocparser.AssocParserConfig()
        self.report = assocparser.Report(group=group, dataset=dataset, config=self.config)
        # self.gpi = None
        if self.config.gpi_authority_path is not None:
            self.bio_entities.merge(collections.BioEntities.load_from_file(self.config.gpi_authority_path))
            print("Loaded {} entities from {}".format(len(self.bio_entities.entities.keys()), self.config.gpi_authority_path))

    def gaf_version(self) -> str:
        if self.version:
            return self.version
        else:
            return self.default_version

    def qualifier_parser(self) -> assocparser.ColumnValidator:
        if self.gaf_version() == "2.2":
            return assocparser.Qualifier2_2()

        return assocparser.Qualifier2_1()

    def skim(self, file):
        file = self._ensure_file(file)
        tuples = []
        for line in file:
            if line.startswith("!"):
                continue
            vals = line.split("\t")
            if len(vals) < 15:
                logger.error("Unexpected number of vals: {}. GAFv1 has 15, GAFv2 has 17.".format(vals))

            split_line = assocparser.SplitLine(line=line, values=vals, taxon=vals[12])

            negated, relation, _ = self._parse_qualifier(vals[3], vals[8])

            # never include NOTs in a skim
            if negated:
                continue
            if self._is_exclude_relation(relation):
                continue
            id = self._pair_to_id(vals[0], vals[1])
            if not self._validate_id(id, split_line, context=ENTITY):
                continue
            n = vals[2]
            t = vals[4]
            tuples.append( (id,n,t) )
        return tuples

    def make_internal_cell_component_closure(self):
        if self.config.ontology:
            self.cell_component_descendants_closure = protein_complex_sublcass_closure(self.config.ontology)

    def parse_line(self, line):
        """
        Parses a single line of a GAF

        Return a tuple `(processed_line, associations)`. Typically
        there will be a single association, but in some cases there
        may be none (invalid line) or multiple (disjunctive clause in
        annotation extensions)

        Note: most applications will only need to call this directly if they require fine-grained control of parsing. For most purposes,
        :method:`parse_file` can be used over the whole file

        Arguments
        ---------
        line : str
            A single tab-seperated line from a GAF file

        """

        # Returns assocparser.ParseResult
        parsed = super().validate_line(line)
        if parsed:
            return parsed

        if self.is_header(line):
            # Save off version info here
            if self.version is None:
                # We are still looking
                parsed = parser_version_regex.findall(line)
                if len(parsed) == 1:
                    filetype, version, _ = parsed[0]
                    if version == "2.2":
                        logger.info("Detected GAF version 2.2")
                        self.version = version
                    else:
                        logger.info("Detected GAF version {}, so using 2.1".format(version))
                        self.version = self.default_version
                        # Compute the cell component subclass closure
                        self.make_internal_cell_component_closure()

            return assocparser.ParseResult(line, [{ "header": True, "line": line.strip() }], False)

        # At this point, we should have gone through all the header, and a version number should be established
        if self.version is None:
            logger.warning("No version number found for this file so we will assume GAF version: {}".format(self.default_version))
            self.version = self.default_version
            self.make_internal_cell_component_closure()

        vals = [el.strip() for el in line.split("\t")]

        # GAF v1 is defined as 15 cols, GAF v2 as 17.
        # We treat everything as GAF2 by adding two blank columns.
        # TODO: check header metadata to see if columns corresponds to declared dataformat version

        parsed = to_association(list(vals), report=self.report, qualifier_parser=self.qualifier_parser(), bio_entities=self.bio_entities)
        if parsed.associations == []:
            return parsed

        assoc = parsed.associations[0]

        # Qualifier is index 3
        # If we are 2.1, and qualifier has no relation
        # Also must have an ontology
        # Then upgrade
        # For https://github.com/geneontology/go-site/issues/1558
        if self.gaf_version() == "2.1" and (vals[3] == "" or vals[3] == "NOT") and self.config.ontology:
            assoc = self.upgrade_empty_qualifier(assoc)

        ## Run GO Rules, save split values into individual variables
        go_rule_results = qc.test_go_rules(assoc, self.config, group=self.group)
        for rule, result in go_rule_results.all_results.items():
            if result.result_type == qc.ResultType.WARNING:
                self.report.warning(line, assocparser.Report.VIOLATES_GO_RULE, "",
                                    msg="{id}: {message}".format(id=rule.id, message=result.message), rule=int(rule.id.split(":")[1]))

            if result.result_type == qc.ResultType.ERROR:
                self.report.error(line, assocparser.Report.VIOLATES_GO_RULE, "",
                                    msg="{id}: {message}".format(id=rule.id, message=result.message), rule=int(rule.id.split(":")[1]))
                # Skip the annotation
                return assocparser.ParseResult(line, [], True)

            if result.result_type == qc.ResultType.PASS:
                self.report.message(assocparser.Report.INFO, line, Report.RULE_PASS, "",
                                    msg="Passing Rule", rule=int(rule.id.split(":")[1]))

        assoc = go_rule_results.annotation  # type: association.GoAssociation
        split_line = assocparser.SplitLine(line=line, values=vals, taxon=str(assoc.object.taxon))

        if self.config.group_idspace is not None and assoc.provided_by not in self.config.group_idspace:
            self.report.warning(line, Report.INVALID_ID, assoc.provided_by,
                "GORULE:0000027: assigned_by is not present in groups reference", taxon=str(assoc.object.taxon), rule=27)

        db = assoc.subject.id.namespace
        if self.config.entity_idspaces is not None and db not in self.config.entity_idspaces:
            # Are we a synonym?
            upgrade = self.config.entity_idspaces.reverse(db)
            if upgrade is not None:
                # If we found a synonym
                self.report.warning(line, Report.INVALID_ID_DBXREF, db, "GORULE:0000027: {} is a synonym for the correct ID {}, and has been updated".format(db, upgrade), taxon=str(assoc.object.taxon), rule=27)
                assoc.subject.id.namespace = upgrade

        ## --
        ## db + db_object_id. CARD=1
        ## --assigned_by
        if not self._validate_id(str(assoc.subject.id), split_line, allowed_ids=self.config.entity_idspaces):
            return assocparser.ParseResult(line, [], True)

        # Using a given gpi file to validate the gene object
        # if self.gpi is not None:
        #     entity = self.gpi.get(str(assoc.subject.id), None)
        #     if entity is not None:
        #         assoc.subject.label = entity["symbol"]
        #         assoc.subject.fullname = entity["name"]
        #         assoc.subject.synonyms = entity["synonyms"].split("|")
        #         assoc.subject.type = entity["type"]

        if not self._validate_id(str(assoc.object.id), split_line, context=ANNOTATION):
            print("skipping because {} not validated!".format(assoc.object.id))
            return assocparser.ParseResult(line, [], True)

        valid_goid = self._validate_ontology_class_id(str(assoc.object.id), split_line)
        if valid_goid is None:
            return assocparser.ParseResult(line, [], True)
        assoc.object.id = association.Curie.from_str(valid_goid)

        references = self.validate_curie_ids(assoc.evidence.has_supporting_reference, split_line)
        if references is None:
            # Reporting occurs in above function call
            return assocparser.ParseResult(line, [], True)

        # With/From
        for wf in assoc.evidence.with_support_from:
            validated = self.validate_curie_ids(wf.elements, split_line)
            if validated is None:
                return assocparser.ParseResult(line, [], True)

        # validation
        self._validate_symbol(assoc.subject.label, split_line)


        ## --
        ## taxon CARD={1,2}
        ## --
        ## if a second value is specified, this is the interacting taxon
        ## We do not use the second value
        valid_taxon = self._validate_taxon(str(assoc.object.taxon), split_line)
        valid_interacting = self._validate_taxon(str(assoc.interacting_taxon), split_line) if assoc.interacting_taxon else True
        if not valid_taxon:
            self.report.error(line, assocparser.Report.INVALID_TAXON, str(assoc.object.taxon), "Taxon ID is invalid", rule=27)
        if not valid_interacting:
            self.report.error(line, assocparser.Report.INVALID_TAXON, str(assoc.interacting_taxon), "Taxon ID is invalid", rule=27)
        if (not valid_taxon) or (not valid_interacting):
            return assocparser.ParseResult(line, [], True)

        return assocparser.ParseResult(line, [assoc], False, vals[6])

    def upgrade_empty_qualifier(self, assoc: association.GoAssociation) -> association.GoAssociation:
        """
        From https://github.com/geneontology/go-site/issues/1558

        For GAF 2.1 we will apply an algorithm to find a best fit relation if the qualifier column is empty.
        If the qualifiers field is empty, then:
            If the GO Term is exactly GO:008150 Biological Process, then the qualifier should be `involved_in`
            If the GO Term is exactly GO:0008372 Cellular Component, then the qualifer should be `is_active_in`
            If the GO Term is a Molecular Function, then the new qualifier should be `enables`
            If the GO Term is a Biological Process, then the new qualifier should be `acts_upstream_or_within
            Otherwise for Cellular Component, if it's subclass of anatomical structure, than use `located_in`
                and if it's a protein-containing complexes, use `part_of`
        :param assoc: GoAssociation
        :return: the possibly upgraded GoAssociation
        """
        term = str(assoc.object.id)
        namespace = self.config.ontology.obo_namespace(term)

        if term == "GO:0008150":
            involved_in = association.Curie(namespace="RO", identity="0002331")
            assoc.qualifiers = [involved_in]
            assoc.relation = involved_in
        elif term == "GO:0008372":
            is_active_in = association.Curie(namespace="RO", identity="0002432")
            assoc.qualifiers = [is_active_in]
            assoc.relation = is_active_in
        elif namespace == "molecular_function":
            enables = association.Curie(namespace="RO", identity="0002327")
            assoc.qualifiers = [enables]
            assoc.relation = enables
        elif namespace == "biological_process":
            acts_upstream_or_within = association.Curie(namespace="RO", identity="0002264")
            assoc.qualifiers = [acts_upstream_or_within]
            assoc.relation = acts_upstream_or_within
        elif namespace == "cellular_component":
            if term in self.cell_component_descendants_closure:
                part_of = association.Curie(namespace="BFO", identity="0000050")
                assoc.qualifiers = [part_of]
                assoc.relation = part_of
            else:
                located_in = association.Curie(namespace="RO", identity="0001025")
                assoc.qualifiers = [located_in]
                assoc.relation = located_in

        self.report.warning(assoc.source_line, Report.INVALID_QUALIFIER,
                            "EMPTY", "GORULE:0000059 Upgrading qualifier/relation to {} when reading GAF 2.1".format(assoc.relation),
                            taxon=str(assoc.subject.taxon), rule=59)
        return assoc

ecomap = EcoMap()
ecomap.mappings()

def to_association(gaf_line: List[str], report=None, group="unknown", dataset="unknown", qualifier_parser=assocparser.Qualifier2_1(), bio_entities=None) -> assocparser.ParseResult:
    report = Report(group=group, dataset=dataset) if report is None else report
    bio_entities = collections.BioEntities(dict()) if bio_entities is None else bio_entities
    source_line = "\t".join(gaf_line)

    if source_line == "":
        report.error(source_line, "Blank Line", "EMPTY", "Blank lines are not allowed", rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    if len(gaf_line) > 17:
        # If we see more than 17 columns, we will just cut off the columns after column 17
        report.warning(source_line, assocparser.Report.WRONG_NUMBER_OF_COLUMNS, "",
            msg="There were more than 17 columns in this line. Proceeding by cutting off extra columns after column 17.",
            rule=1)
        gaf_line = gaf_line[:17]

    if 17 > len(gaf_line) >= 15:
        gaf_line += [""] * (17 - len(gaf_line))

    if len(gaf_line) != 17:
        report.error(source_line, assocparser.Report.WRONG_NUMBER_OF_COLUMNS, "",
            msg="There were {columns} columns found in this line, and there should be 15 (for GAF v1) or 17 (for GAF v2)".format(columns=len(gaf_line)), rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    ## check for missing columns
    ## We use indeces here because we run GO RULES before we split the vals into individual variables
    DB_INDEX = 0
    DB_OBJECT_INDEX = 1
    TAXON_INDEX = 12
    REFERENCE_INDEX = 5
    if gaf_line[DB_INDEX] == "":
        report.error(source_line, Report.INVALID_IDSPACE, "EMPTY", "col1 is empty", taxon=gaf_line[TAXON_INDEX], rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)
    if gaf_line[DB_OBJECT_INDEX] == "":
        report.error(source_line, Report.INVALID_ID, "EMPTY", "col2 is empty", taxon=gaf_line[TAXON_INDEX], rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)
    if gaf_line[REFERENCE_INDEX] == "":
        report.error(source_line, Report.INVALID_ID, "EMPTY", "reference column 6 is empty", taxon=gaf_line[TAXON_INDEX], rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    parsed_taxons_result = gaf_line_validators["taxon"].validate(gaf_line[TAXON_INDEX])  # type: assocparser.ValidateResult
    if not parsed_taxons_result.valid:
        report.error(source_line, Report.INVALID_TAXON, parsed_taxons_result.original, parsed_taxons_result.message, taxon=parsed_taxons_result.original, rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    taxon = parsed_taxons_result.parsed[0]

    date = assocparser.parse_date(gaf_line[13], report, source_line)
    if date is None:
        return assocparser.ParseResult(source_line, [], True, report=report)

    interacting_taxon = parsed_taxons_result.parsed[1] if len(parsed_taxons_result.parsed) == 2 else None
    subject_curie = association.Curie(gaf_line[0], gaf_line[1])
    subject = association.Subject(subject_curie, gaf_line[2], [gaf_line[9]], gaf_line[10].split("|"), [association.map_gp_type_label_to_curie(gaf_line[11])], taxon)
    gpi_entity = bio_entities.get(subject_curie)
    if gpi_entity is not None and subject != gpi_entity:
        subject = gpi_entity

    aspect = gaf_line[8]
    negated, relation_label, qualifiers = assocparser._parse_qualifier(gaf_line[3], aspect)
    # Note: Relation label is grabbed from qualifiers, if any exist in _parse_qualifier
    qualifiers = [association.Curie.from_str(curie_util.contract_uri(relations.lookup_label(q), strict=False)[0]) for q in qualifiers]

    # column 4 is qualifiers -> index 3
    # For allowed, see http://geneontology.org/docs/go-annotations/#annotation-qualifiers
    # We use the below validate to check validaty if qualifiers, not as much to *parse* them into the GoAssociation object.
    # For GoAssociation we will use the above qualifiers list. This is fine because the above does not include `NOT`, etc
    # This is confusing, and we can fix later on by consolidating qualifier and relation in GoAssociation.
    parsed_qualifiers = qualifier_parser.validate(gaf_line[3])
    if not parsed_qualifiers.valid:
        report.error(source_line, Report.INVALID_QUALIFIER, parsed_qualifiers.original, parsed_qualifiers.message, taxon=gaf_line[TAXON_INDEX], rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    object = association.Term(association.Curie.from_str(gaf_line[4]), taxon)
    if isinstance(object, association.Error):
        report.error(source_line, Report.INVALID_SYMBOL, gaf_line[4], "Problem parsing GO Term", taxon=gaf_line[TAXON_INDEX], rule=1)

    # References
    references = [association.Curie.from_str(e) for e in gaf_line[5].split("|") if e]
    for r in references:
        if isinstance(r, association.Error):
            report.error(source_line, Report.INVALID_SYMBOL, gaf_line[5], "Problem parsing references", taxon=gaf_line[TAXON_INDEX], rule=1)
            return assocparser.ParseResult(source_line, [], True, report=report)

    gorefs = [ref for ref in references if ref.namespace == "GO_REF"] + [None]
    eco_curie = ecomap.coderef_to_ecoclass(gaf_line[6], reference=gorefs[0])
    if eco_curie is None:
        report.error(source_line, Report.UNKNOWN_EVIDENCE_CLASS, gaf_line[6], msg="Expecting a known ECO GAF code, e.g ISS", rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    withfroms = association.ConjunctiveSet.str_to_conjunctions(gaf_line[7])
    if isinstance(withfroms, association.Error):
        report.error(source_line, Report.INVALID_SYMBOL, gaf_line[7], "Problem parsing with/from", taxon=gaf_line[TAXON_INDEX], rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    evidence_type = association.Curie.from_str(eco_curie)
    if isinstance(evidence_type, association.Error):
        report.error(source_line, Report.INVALID_SYMBOL, gaf_line[6], "Problem parsing evidence type", taxon=gaf_line[TAXON_INDEX], rule=1)

    evidence = association.Evidence(association.Curie.from_str(eco_curie), references, withfroms)
    if any([isinstance(e, association.Error) for e in evidence.has_supporting_reference]):
        first_error = [e for e in evidence.has_supporting_reference if isinstance(e, association.Error)][0]
        report.error(source_line, Report.INVALID_SYMBOL, gaf_line[5], first_error.info, taxon=str(taxon), rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    subject_extensions = []
    if gaf_line[16]:
        subject_filler = association.Curie.from_str(gaf_line[16])
        if isinstance(subject_filler, association.Error):
            report.error(source_line, assocparser.Report.INVALID_ID, gaf_line[16], subject_filler.info, taxon=str(taxon), rule=1)
            return assocparser.ParseResult(source_line, [], True, report=report)
        # filler is not an Error, so keep moving
        subject_extensions.append(association.ExtensionUnit(association.Curie.from_str("rdfs:subClassOf"), subject_filler))

    conjunctions = []
    if gaf_line[15]:
        conjunctions = association.ConjunctiveSet.str_to_conjunctions(
            gaf_line[15],
            conjunct_element_builder=lambda el: association.ExtensionUnit.from_str(el))

        if isinstance(conjunctions, association.Error):
            report.error(source_line, Report.EXTENSION_SYNTAX_ERROR, conjunctions.info, "extensions should be relation(curie) and relation should have corresponding URI", taxon=str(taxon), rule=1)
            return assocparser.ParseResult(source_line, [], True, report=report)

    relation_uri = relations.lookup_label(relation_label)
    if relation_uri is None:
        report.error(source_line, assocparser.Report.INVALID_QUALIFIER, relation_label, "Could not find CURIE for relation `{}`".format(relation_label), taxon=str(taxon), rule=1)
        return assocparser.ParseResult(source_line, [], True, report=report)

    # We don't have to check that this is well formed because we're grabbing it from the known relations URI map.
    relation_curie = association.Curie.from_str(curie_util.contract_uri(relation_uri)[0])

    a = association.GoAssociation(
        source_line="\t".join(gaf_line),
        subject=subject,
        relation=relation_curie,
        object=object,
        negated=negated,
        qualifiers=qualifiers,
        aspect=aspect,
        interacting_taxon=interacting_taxon,
        evidence=evidence,
        subject_extensions=subject_extensions,
        object_extensions=conjunctions,
        provided_by=gaf_line[14],
        date=date,
        properties={})

    return assocparser.ParseResult(source_line, [a], False, report=report)



import pytest

from ontobio.model import association
from ontobio.model.association import Curie
from ontobio.io import gafparser

def test_negated_qualifiers():

    gaf = ["PomBase", "SPBC11B10.09", "cdc2", "NOT", "GO:0007275", "PMID:21873635", "IBA", "PANTHER:PTN000623979|TAIR:locus:2099478", "P", "Cyclin-dependent kinase 1", "UniProtKB:P04551|PTN000624043", "protein", "taxon:284812", "20170228", "GO_Central", "", ""]
    association = gafparser.to_association(gaf).associations[0]
    to_gaf = association.to_gaf_2_1_tsv()
    assert to_gaf[3] == "NOT"

    to_gpad = association.to_gpad_1_2_tsv()
    assert to_gpad[2] == "NOT|involved_in"

def test_conjunctiveset_tostring():
    c = association.ConjunctiveSet(["MGI:12345"])
    assert str(c) == "MGI:12345"

    c = association.ConjunctiveSet(["MGI:12345", "MGI:12345"])
    assert str(c) == "MGI:12345,MGI:12345"

    c = association.ConjunctiveSet([])
    assert str(c) == ""

def test_conjunctive_set_str_to_conjunctions():
    c = association.ConjunctiveSet.str_to_conjunctions("")
    assert c == []

    c = association.ConjunctiveSet.str_to_conjunctions("MGI:12345")
    assert c == [association.ConjunctiveSet([Curie.from_str("MGI:12345")])]

    c = association.ConjunctiveSet.str_to_conjunctions("MGI:12345,MGI:12345")
    assert c == [association.ConjunctiveSet([Curie.from_str("MGI:12345"), Curie.from_str("MGI:12345")])]

    c = association.ConjunctiveSet.str_to_conjunctions("MGI:12345|MGI:12345")
    assert c == [association.ConjunctiveSet([Curie.from_str("MGI:12345")]), association.ConjunctiveSet([Curie.from_str("MGI:12345")])]

    c = association.ConjunctiveSet.str_to_conjunctions("MGI:12345,DOI:333|GO:987")
    assert c == [
            association.ConjunctiveSet([Curie.from_str("MGI:12345"), Curie.from_str("DOI:333")]),
            association.ConjunctiveSet([Curie.from_str("GO:987")])
        ]

    c = association.ConjunctiveSet.str_to_conjunctions("part_of(MGI:123),part_of(FOO:456)|has_direct_input(BAR:987)", conjunct_element_builder=lambda u: association.ExtensionUnit.from_str(u))
    assert c == [
            association.ConjunctiveSet([association.ExtensionUnit(Curie.from_str("BFO:0000050"), Curie.from_str("MGI:123")), association.ExtensionUnit(Curie.from_str("BFO:0000050"), Curie.from_str("FOO:456"))]),
            association.ConjunctiveSet([association.ExtensionUnit(Curie.from_str("GOREL:0000752"), Curie.from_str("BAR:987"))])
        ]

    c = association.ConjunctiveSet.str_to_conjunctions("HECK_NO,part_of(FOO:456)|has_direct_input(BAR:987)", conjunct_element_builder=lambda u: association.ExtensionUnit.from_str(u))
    assert c == association.Error("HECK_NO")

def test_conjunctive_set_list_to_str():
    c = association.ConjunctiveSet.list_to_str([])
    assert c == ""

    c = association.ConjunctiveSet.list_to_str([association.ConjunctiveSet([Curie.from_str("MGI:12345")])])
    assert c == "MGI:12345"

    c = association.ConjunctiveSet.list_to_str([association.ConjunctiveSet([Curie.from_str("MGI:12345"), Curie.from_str("MGI:12345")])])
    assert c == "MGI:12345,MGI:12345"

    c = association.ConjunctiveSet.list_to_str([association.ConjunctiveSet([Curie.from_str("MGI:12345")]), association.ConjunctiveSet([Curie.from_str("MGI:12345")])])
    assert c == "MGI:12345|MGI:12345"

    c = association.ConjunctiveSet.list_to_str([
            association.ConjunctiveSet([Curie.from_str("MGI:12345"), Curie.from_str("DOI:333")]),
            association.ConjunctiveSet([Curie.from_str("GO:987")])
        ])
    assert c == "MGI:12345,DOI:333|GO:987"

def test_date():
    date_str = "20210105"
    date = association.Date(date_str[0:4], date_str[4:6], date_str[6:8], "")
    assert association.ymd_str(date, "-") == "2021-01-05"

def test_subject_fullname():
    s = association.Subject(Curie.from_str("HELLO:12345"), "Hello object", ["fullname"], [], ["protein"], association.Curie.from_str("NCBITaxon:12345"))
    assert "fullname" == s.fullname_field()

    s = association.Subject(Curie.from_str("HELLO:12345"), "Hello object", ["fullname", "another_name"], [], ["protein"], association.Curie.from_str("NCBITaxon:12345"))
    assert "fullname|another_name" == s.fullname_field()

    s = association.Subject(Curie.from_str("HELLO:12345"), "Hello object", ["fullname"], [], ["protein"], association.Curie.from_str("NCBITaxon:12345"))
    assert "fullname" == s.fullname_field(max=1)

def test_subject_types_unknown():
    s = association.Subject(Curie.from_str("HELLO:12345"), "Hello object", ["fullname"], [], [], association.Curie.from_str("NCBITaxon:12345"))
    assert [Curie(namespace="CHEBI", identity="33695")] == s.type

def test_subject_types_label():
    s = association.Subject(Curie.from_str("HELLO:12345"), "Hello object", ["fullname"], [], ["mRNA", "tRNA"], association.Curie.from_str("NCBITaxon:12345"))
    assert [Curie.from_str("SO:0000234"), Curie.from_str("SO:0000253")] == s.type


import gzip
import click
import os

from functools import wraps

def gzips(file_function):

    @wraps(file_function)
    def wrapper(*args, **kwargs):
        output_file = file_function(*args, **kwargs)
        if isinstance(output_file, list):
            for o in output_file:
                zipup(o)
        else:
            zipup(output_file)

        return output_file

    return wrapper

def zipup(file_path):
    click.echo("Zipping {}".format(file_path))
    path, filename = os.path.split(file_path)
    zipname = "{}.gz".format(filename)
    target = os.path.join(path, zipname)

    with open(file_path, "rb") as p:
        with gzip.open(target, "wb") as tf:
            tf.write(p.read())
            
def unzip(path, target):
    click.echo("Unzipping {}".format(path))
    def chunk_gen():
        with gzip.open(path, "rb") as p:
            while True:
                chunk = p.read(size=512 * 1024)
                if not chunk:
                    break
                yield chunk

    with open(target, "wb") as tf:
        with click.progressbar(iterable=chunk_gen()) as chunks:
            for chunk in chunks:
                tf.write(chunk)

def find(l, finder):
    filtered = [n for n in l if finder(n)]
    if len(filtered) == 0:
        return None
    else:
        return filtered[0]



import csv
import logging
from typing import List
from ontobio.ontol import Ontology

def expand_tsv(input: str, ontology: Ontology = None, outfile=None, sep='\t', cols: List[str]=None) -> None:
    """
    Adds additional columns to a TSV by performing additional ontology lookups

    For example, given a TSV with a column `term`, this can add a column `term_label`
    in future it may also add closures

    :param input: filename of a TSV (must have column headers)
    :param ontology: used for lookup
    :param outfile: pathname for output file
    :param sep: delimiter
    :param cols: names of columns
    :return:
    """
    with open(input, newline='') as io:
        reader = csv.DictReader(io, delimiter='\t')
        items = []
        if True:
            outwriter = csv.writer(outfile, delimiter=sep)
            first = True
            for row in reader:
                if first:
                    first = False
                    hdr = []
                    for k in row.keys():
                        hdr.append(k)
                        if k in cols:
                            hdr.append(f'{k}_label')
                    outwriter.writerow(hdr)
                vals = []
                for k,v in row.items():
                    vals.append(v)
                    if k in cols:
                        id = row[k]
                        label = ontology.label(id)
                        vals.append(label)
                        if label is None:
                            logging.warning(f"No id: {id}")
                        #item[f'{k}_label'] = label
                outwriter.writerow(vals)




from ontobio.model import collections
from ontobio.model.association import Curie, Subject
from ontobio.io import assocparser, gafparser, gpadparser

def test_bioentities_get_when_empty():
    e = collections.BioEntities(dict())
    assert e.get(Curie("FOO", "123")) == None

def test_bioentities_get_when_exists():
    e = collections.BioEntities({
        Curie("FOO", "123"): Subject(Curie("FOO", "123"), "hello", "world", [], "protien", Curie("NCBITaxon", "12345"))
    })
    assert e.get(Curie.from_str("FOO:123")) == Subject(Curie("FOO", "123"), "hello", "world", [], "protien", Curie("NCBITaxon", "12345"))

def test_bioentities_merge():
    e = collections.BioEntities({
        Curie("FOO", "123"): Subject(Curie("FOO", "123"), "hello", "world", [], "protien", Curie("NCBITaxon", "12345"))
    })

    o = collections.BioEntities({
        Curie("BAR", "987"): Subject(Curie("BAR", "987"), "goodbye", "world", [], "protien", Curie("NCBITaxon", "999"))
    })

    assert e.merge(o) == collections.BioEntities({
        Curie("FOO", "123"): Subject(Curie("FOO", "123"), "hello", "world", [], "protien", Curie("NCBITaxon", "12345")),
        Curie("BAR", "987"): Subject(Curie("BAR", "987"), "goodbye", "world", [], "protien", Curie("NCBITaxon", "999"))
    })

def test_bioentities_merge_clobber():
    e = collections.BioEntities({
        Curie("FOO", "123"): Subject(Curie("FOO", "123"), "hello", "world", [], "protien", Curie("NCBITaxon", "12345"))
    })

    o = collections.BioEntities({
        Curie("FOO", "123"): Subject(Curie("FOO", "123"), "different", "world", [], "dog", Curie("NCBITaxon", "12345"))
    })
    # Get the clobbered key, the value should be the subject in `o`
    assert e.merge(o).get(Curie("FOO", "123")) == Subject(Curie("FOO", "123"), "different", "world", [], "dog", Curie("NCBITaxon", "12345"))

def test_bioentities_load_from_file():
    pombase = collections.BioEntities.load_from_file("tests/resources/truncated-pombase.gpi")
    assert len(pombase.entities.keys()) == 199 # Has 199 gpi lines in the file
    assert pombase.get(Curie.from_str("PomBase:SPAC1565.04c")) == Subject(Curie.from_str("PomBase:SPAC1565.04c"), "ste4", ["adaptor protein Ste4"], [], ["protein"], Curie.from_str("NCBITaxon:4896"))

def test_bioentities_file_gone():
    gone = collections.BioEntities.load_from_file("not_here.gpi")
    assert gone.entities == dict()

def test_create_parser_non_version_line():
    parser = collections.create_parser_from_header("! hello", assocparser.AssocParserConfig())
    assert parser == None

def test_create_parser_gaf():
    parser = collections.create_parser_from_header("!gaf-version: 2.1", assocparser.AssocParserConfig())
    assert isinstance(parser, gafparser.GafParser)
    assert parser.version == "2.1"

    parser = collections.create_parser_from_header("!gaf-version: 2.2", assocparser.AssocParserConfig())
    assert isinstance(parser, gafparser.GafParser)
    assert parser.version == "2.2"

def test_create_parser_gpad():
    parser = collections.create_parser_from_header("!gpa-version: 1.2", assocparser.AssocParserConfig())
    assert isinstance(parser, gpadparser.GpadParser)
    assert parser.version == "1.2"

    parser = collections.create_parser_from_header("!gpad-version: 2.0", assocparser.AssocParserConfig())
    assert isinstance(parser, gpadparser.GpadParser)
    assert parser.version == "2.0"

def test_construct_collection_empty():
    collection = collections.construct_collection(None, [], assocparser.AssocParserConfig())
    assert collection.headers == []
    assert collection.associations == collections.GoAssociations([])
    assert collection.entities == collections.BioEntities(dict())

def test_construct_collection_basic():
    collection = collections.construct_collection("tests/resources/truncated-pombase.gaf", ["tests/resources/truncated-pombase.gpi"], assocparser.AssocParserConfig())
    # There are 24 lines of headers in the truncated-pombase.gaf
    assert len(collection.headers) == 25
    # There are many lines in this file, we should get more than 100 back
    assert len(collection.associations.associations) > 100
    # This is a 199 line gpi file, with each identifier unique, so should have 200 entity entries
    assert len(collection.entities.entities.keys()) == 199

def test_construct_collection_no_version_error():
    collection = collections.construct_collection("tests/resources/no-version.gaf", [], assocparser.AssocParserConfig())
    assert collection.associations.associations == []
    assert collection.report.to_report_json()["messages"]["gorule-0000001"][0]["type"] == "Invalid Annotation File"

def test_bioentities_from_gpi_2_0():
    entities = collections.BioEntities.load_from_file("tests/resources/mgi.truncated.gpi2")
    assert entities.get(Curie(namespace="MGI", identity="MGI:1918925")) == Subject(
        id=Curie.from_str("MGI:MGI:1918925"),
        label="0610010F05Rik",
        fullname=["RIKEN cDNA 0610010F05 gene"],
        synonyms=[], type=[Curie.from_str("SO:0001217")],
        taxon=Curie(namespace="NCBITaxon", identity="10090"),
        db_xrefs=[Curie.from_str("UniProtKB:Q68FF0")])



import pytest
from ontobio.io import assocparser
from ontobio.io.gpadparser import to_association
from ontobio.ontol_factory import OntologyFactory
from ontobio.rdfgen.gocamgen import collapsed_assoc, gocam_builder, gocamgen

GO_ONTO = OntologyFactory().create("tests/resources/go-binding.json")

def test_evidence_max_date():
    ev1 = gocamgen.GoCamEvidence(code="ECO:0000314", references=["PMID:12345"], date="2008-08-14")
    ev2 = gocamgen.GoCamEvidence(code="ECO:0000314", references=["PMID:12345"], date="2011-04-12")
    ev3 = gocamgen.GoCamEvidence(code="ECO:0000314", references=["PMID:12345"], date="2021-03-01")
    max_date = gocamgen.GoCamEvidence.max_date([ev1, ev2, ev3])
    assert max_date == "2021-03-01"

def test_get_with_froms():
    gpi_ents = gocam_builder.GoCamBuilder.parse_gpi(gpi_file="tests/resources/mgi2.test_entities.gpi")
    report = assocparser.Report(group="unknown", dataset="unknown")
    vals = [
        "MGI:MGI:1929608",
        "",
        "RO:0001025",
        "GO:0005515",
        "MGI:MGI:3028747|PMID:14627703",
        "ECO:0000353",
        "MGI:MGI:1915834",  # with/from
        "",
        "2004-04-30",
        "MGI",
        "",
        "creation-date=2004-04-30|modification-date=2004-04-30|contributor-id=http://orcid.org/0000-0002-9796-7693"
    ]
    result = to_association(list(vals), report=report, version="2.0")
    go_assoc = result.associations[0]
    ca_set = collapsed_assoc.CollapsedAssociationSet(GO_ONTO, gpi_ents)
    with_froms = ca_set.get_with_froms(go_assoc)
    assert len(with_froms) == 1 and \
           with_froms[0].header == ["MGI:MGI:1915834"] and \
           with_froms[0].line == []

    vals[6] = "MGI:MGI:1915834|FAKE:12345"
    result = to_association(list(vals), report=report, version="2.0")
    go_assoc = result.associations[0]
    ca_set = collapsed_assoc.CollapsedAssociationSet(GO_ONTO, gpi_ents)
    with_froms = ca_set.get_with_froms(go_assoc)
    # FAKE:12345 should be on line since not in GPI nor does it have same taxon as subject MGI:MGI:1929608
    assert len(with_froms) == 2 and \
           with_froms[0].header == ["MGI:MGI:1915834"] and \
           with_froms[0].line == [] and \
           with_froms[1].header == [] and \
           with_froms[1].line == ["FAKE:12345"]

    # Test merging of same-header with/from values in different order
    ca_set = collapsed_assoc.CollapsedAssociationSet(GO_ONTO, gpi_ents)
    header1 = collapsed_assoc.GoAssocWithFrom(header=["MGI:MGI:1915834", "FAKE:12345"])
    header2 = collapsed_assoc.GoAssocWithFrom(header=["FAKE:12345", "MGI:MGI:1915834"])
    ca_set.find_or_create_collapsed_association(go_assoc, with_from=header1)
    ca_set.find_or_create_collapsed_association(go_assoc, with_from=header2)
    assert len(ca_set.collapsed_associations) == 1 and ca_set.collapsed_associations[0].with_froms == ["FAKE:12345", "MGI:MGI:1915834"]



# from ontobio.io.gafparser import GafParser
from ontobio.io.gpadparser import GpadParser
from ontobio.io.assocparser import SplitLine
from ontobio.ontol_factory import OntologyFactory
# from prefixcommons import curie_util
from ontobio.ecomap import EcoMap
from ontobio.util.go_utils import GoAspector
from ontobio.rdfgen.assoc_rdfgen import prefix_context
from .filter_rule import *
from .collapsed_assoc import extract_properties
import json
import csv
import os
import argparse
import logging
import datetime
from pathlib import Path


DISTINCT_EXTENSIONS = {}

# logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)
logger.setLevel("INFO")

parser = argparse.ArgumentParser()
parser.add_argument("-f", "--filename")
parser.add_argument("-d", "--dir")
parser.add_argument("-c", "--constraints_yaml")
parser.add_argument("-o", "--out_file")
parser.add_argument("-l", "--leftovers_out_file")
parser.add_argument("-p", "--pattern")
parser.add_argument("-q", "--pattern_outfile")
parser.add_argument("-r", "--pattern_sourcefile")
parser.add_argument("-m", "--mod")
parser.add_argument("-e", "--extensions_list", action='store_const', const=True,
                    help="Print out distinct extensions list")

ontology_prefixes = []
for k, v in prefix_context.items():
    if v.startswith("http://purl.obolibrary.org/obo/"):
        ontology_prefixes.append(k)
ontology_prefixes.append("Pfam")

mod_prefixes = [
    "PomBase",
    "UniProtKB",
]

acceptable_evidence_codes = [
    "EXP",
    "IDA",
    "IPI",
    "IMP",
    "IGI",
    "IEP"
]
ecomap = EcoMap()


def date_fname(fname):
    date_str = datetime.date.today().isoformat()
    fname_and_ext = os.path.splitext(fname)
    new_fname = fname_and_ext[0]
    new_fname += "_{}".format(date_str.replace("-", ""))
    new_fname += fname_and_ext[1]
    return new_fname

class ExtRelationPattern:
    def __init__(self, ext_relation, ext_namespaces, primary_term_roots=None, max_allowed=None):
        self.ext_relation = ext_relation
        self.ext_namespaces = ext_namespaces
        self.max_allowed = max_allowed  # cardinality?
        self.primary_term_roots = primary_term_roots if isinstance(primary_term_roots, list) else [primary_term_roots]
        # TODO: Also handle primary_term_roots == None


class ExtRelationValidPattern(ExtRelationPattern):
    def __init__(self, ext_relation, ext_namespaces, primary_term_roots=None, max_allowed=None):
        ExtRelationPattern.__init__(self, ext_relation, ext_namespaces, primary_term_roots, max_allowed)
        self.is_valid = True


class ExtRelationInvalidPattern(ExtRelationPattern):
    def __init__(self, ext_relation, ext_namespaces, primary_term_roots=None, max_allowed=None):
        ExtRelationPattern.__init__(self, ext_relation, ext_namespaces, primary_term_roots, max_allowed)
        self.is_valid = False


# Extension rules - These should live in separate files
FUNCTION_SINGLES_ONLY = [
    "occurs_in(GO:C)",
    "occurs_in(CL)",
    "occurs_in(UBERON)",
    "occurs_in(EMAPA)",
    "has_input(geneID)",
    "has_input(CHEBI)",
    "has_direct_input(geneID)",
    "has_direct_input(CHEBI)",
    "happens_during(GO:P)",
    "part_of(GO:P)",
    "has_regulation_target(geneID)",
    "activated_by(CHEBI)",
    "inhibited_by(CHEBI)"
]
COMPONENT_SINGLES_ONLY = [
    "part_of(GO:C)",
    "part_of(CL)",
    "part_of(UBERON)",
    "part_of(EMAPA)"
]
PROCESS_SINGLES_ONLY = [
    "occurs_in(GO:C)",
    "occurs_in(CL)",
    "occurs_in(UBERON)",
    "occurs_in(EMAPA)",
    "occurs_in(WBbt)",
    "has_input(geneID)",
    "has_input(CHEBI)",
    "has_direct_input(geneID)",
    "has_direct_input(CHEBI)",
    "has_regulation_target(geneID)",
    "part_of(GO:P)"
]
EXTENSION_RELATION_UNIVERSE = []  # Should this also be aspect-specific?
for r in FUNCTION_SINGLES_ONLY + COMPONENT_SINGLES_ONLY + PROCESS_SINGLES_ONLY:
    if r not in EXTENSION_RELATION_UNIVERSE:
        EXTENSION_RELATION_UNIVERSE.append(r)
# Add remaining extensions not having cardinality restrictions yet we know how to translate them
EXTENSION_RELATION_UNIVERSE.append([
    "has_regulation_target(geneID)"
])


gaf_indices = {}
gpad_indices = {

}

# TODO: Check where Aspector is used in gocamgen. If only for validating extensions we may be able to remove it.
class CachedGoAspector(GoAspector):

    def __init__(self, cache_filepath=None, go_ontology=None):
        GoAspector.__init__(self, go_ontology)
        if cache_filepath is None:
            # TODO: Where to store cache?
            cache_filepath = "resources/aspect_lookup.json"
        self.cache_filepath = cache_filepath
        self.aspect_lookup = {
            "F": [],
            "P": [],
            "C": []
        }
        aspect_file = Path(self.cache_filepath)
        if not aspect_file.is_file():
            logger.debug("Creating aspect_lookup cache file: {}".format(self.cache_filepath))
            self.write_cache()
        else:
            with open(cache_filepath) as af:
                try:
                    self.aspect_lookup = json.load(af)
                except json.decoder.JSONDecodeError:
                    logger.warning("Corrupt aspect_lookup cache file: {} - Recreating...".format(self.cache_filepath))

    def write_cache(self):
        with open(self.cache_filepath, "w+") as af:
            json.dump(self.aspect_lookup, af)

    def _is_biological_process(self, go_term):
        bp_root = "GO:0008150"
        return go_term in self.aspect_lookup["P"] or go_term == bp_root

    def _is_molecular_function(self, go_term):
        mf_root = "GO:0003674"
        return go_term in self.aspect_lookup["F"] or go_term == mf_root

    def _is_cellular_component(self, go_term):
        cc_root = "GO:0005575"
        return go_term in self.aspect_lookup["C"] or go_term == cc_root

    def go_aspect(self, term):
        if self._is_molecular_function(term):
            return 'F'
        elif self._is_biological_process(term):
            return 'P'
        elif self._is_cellular_component(term):
            return 'C'
        else:
            aspect = super(CachedGoAspector, self).go_aspect(term)
            if aspect:
                self.aspect_lookup[aspect].append(term)
                return aspect
            return None



def format_extensions(ext_json):
    # print(ext_json)
    ej = json.loads(ext_json)
    ret = []
    # print(ej["union_of"][0]["intersection_of"])
    for i in ej["union_of"][0]["intersection_of"]:
        ret.append("{}({})".format(i["property"], i["filler"]))
    # print(ret)
    return ",".join(ret)

def get_relation_and_term(ext):
    ext_parts = ext.split("(")
    relation = ext_parts[0]  # i["property"]
    try:
        ext_term = ext_parts[1].split(")")[0]  # i["filler"]
    except:
        logger.warning(ext_parts)
        # [' dppa4']
        # $ grep dppa4 resources/mgi.gpa.test.gpa
        # MGI	MGI:2141165	acts_upstream_of_or_within	GO:0010628	MGI:MGI:5913752|PMID:20699224	ECO:0000315	MGI:MGI:6108355		20190124	MGI	has_regulation_target(Dppa2, dppa4, piwil2),occurs_in(EMAPA:16036)
        ext_term = ext_parts[1].split(")")[0]  # i["filler"]
    return relation, ext_term

def filter_evi_codes(annots):
    filtered = []
    for a in annots:
        if ecomap.ecoclass_to_coderef(a[5])[0] in acceptable_evidence_codes:
            filtered.append(a)
    return filtered

def filter_has_extension(annots):
    filtered = []
    for a in annots:
        if a[10] != "":
            filtered.append(a)
    return filtered

def filter_rule_validate_lines(annots, assoc_filter):
    filtered = []
    # Converts split GPAD line into ontobio assoc obj for passing into standard FilterRule validation
    gpad_parser = GpadParser()
    for a in annots:
        parse_result = gpad_parser.parse_line("\t".join(a))
        if len(parse_result.associations) > 0:
            # Right now, GpadParser only returns 0 or 1 associations
            assoc = parse_result.associations[0]
            assoc = extract_properties(assoc)
            if "annotation_properties" in assoc:
                a.append(assoc["annotation_properties"])
            if assoc_filter.validate_line(assoc):
                filtered.append(a)
    return filtered

def sum_combos(extension_counts, combo_list):
    cur_sum = 0
    for c in combo_list:
        if c in extension_counts:
            cur_sum += extension_counts[c]
    return cur_sum

def violates_combo_rule(extension_counts, list_of_combo_lists, max_allowed):
    for combo_list in list_of_combo_lists:
        if sum_combos(extension_counts, combo_list) > max_allowed:
            return True
    return False

def is_nested_occurs_in_grouping(extensions_list):
    # When annotation extensions are comma delimited and are from GO_CC, CL and an anatomy ontology
    # then the model should indicate GO_CC -> part_of->CL->part_of->anatomy.
    # Filter occurs_in extensions
    # Gather terms, arrange
    # When would multiple occurs_in's cause this to fail? Multiple occurs_in(GO:CC), unless CC's are related? Examples?
    #
    do_stuff = 1

class ExtensionValidityCheckResult:
    def __init__(self, is_valid, reason=None, offending_extension=None):
        self.is_valid = is_valid
        self.reason = reason
        self.offending_extension = offending_extension

    def __nonzero__(self):
        return self.is_valid


class ExtensionsMapper:
    def __init__(self, constraints_yaml, ontology=None):
        self.ontology = self.setup_ontologies(ontology)
        self.go_aspector = CachedGoAspector("resources/aspect_lookup.json", go_ontology=go_ontology)

        self.ext_relation_valid_patterns = []
        # Parsing formatted version of David's spreadsheet extensions_list_w_provided_bys_updated_20190328 into more rules
        with open(constraints_yaml) as rf:
            extension_constraints = yaml.load(rf)
            # rel_reader = csv.reader(rf, delimiter="\t")
            # next(rel_reader)  # Skip header row
            for ec in extension_constraints:
                relation = ec["relation"]
                namespaces = ec["namespaces"]
                root_terms = ec["primary_root_terms"]
                max_allowed = ec.get("cardinality")
                # if len(rl[3]) > 0:
                #     max_allowed = int(rl[3])
                ext_pattern = ExtRelationValidPattern(relation, namespaces, root_terms, max_allowed=max_allowed)
                self.ext_relation_valid_patterns.append(ext_pattern)

    @staticmethod
    def setup_ontologies(ontology=None):
        if ontology is None:
            return OntologyFactory().create("go")
        return ontology

    def is_valid_ext_pattern(self, primary_term, ext_relation, ext_namespace, occurence_count):
        for p in self.ext_relation_valid_patterns:
            if p.ext_relation == ext_relation and ext_namespace in p.ext_namespaces and \
                    not (p.max_allowed and occurence_count > p.max_allowed):
                # Check ancestors of query term - might take a while
                all_ancestors = GO_ONTOLOGY.ancestors(primary_term, reflexive=True)
                is_a_ancestors = GO_ONTOLOGY.subontology(all_ancestors).ancestors(primary_term,
                                                                                  relations=["subClassOf"],
                                                                                  reflexive=True)
                # Cache IS_A_ANCESTORS[primary_term] = is_a_ancestors ?
                # Valid pattern root term is an is_a ancestor of GPAD primary term
                for t in p.primary_term_roots:
                    if t in is_a_ancestors:
                        return True
        return False

    def following_rules(self, extension_list, aspect, term):
        ext_counts = {}
        for e in extension_list:
            if e in ext_counts:
                ext_counts[e] += 1
            else:
                ext_counts[e] = 1

        for ek in ext_counts.keys():
            if not self.is_valid_ext_pattern(term, *get_relation_and_term(ek), ext_counts[ek]):
                # return False
                # Pack this into Error class
                return ExtensionValidityCheckResult(is_valid=False, offending_extension=ek)

        combos_to_check_for = [
            ["occurs_in(UBERON)", "occurs_in(EMAPA)"]
        ]
        # Only allow 1 occurrence each of combos_to_check_for
        if violates_combo_rule(ext_counts, combos_to_check_for, 1):
            # return False
            return ExtensionValidityCheckResult(is_valid=False, reason="violates_combo_rule")

        # return True
        return ExtensionValidityCheckResult(is_valid=True)

    def filter_no_rules_broken(self, annots):
        filtered = []
        for a in annots:
            if not self.following_rules(a):
                filtered.append(a)
        return filtered

    def translate_relation_to_ro(self, relation_label):
        for n in self.ontology.nodes():
            node_label = self.ontology.label(n)
            if node_label == relation_label.replace("_", " "):
                return n

    def extensions_list(self, intersection_extensions, row_cols=[]):
        ext_list = []
        # Assuming these extensions are already separated by comma
        intersection_extensions = self.dedupe_extensions(intersection_extensions)
        for i in intersection_extensions:
            relation, ext_term = i['property'], i['filler']
            term_prefix = ext_term.split(":")[0]
            # is prefix for ontology or mod?
            if term_prefix in ontology_prefixes:
                if term_prefix == "GO":
                    # need to find aspect of GO term
                    go_aspect = self.go_aspector.go_aspect(ext_term)
                    if go_aspect is not None:
                        ont = "GO:" + go_aspect
                    else:
                        logger.warning("No aspect found for term: {}".format(ext_term))
                        ont = "GO"
                else:
                    ont = term_prefix
                term_prefix = ont
            else:
                term_prefix = "geneID"
            ext_list.append("{}({})".format(relation, term_prefix))
            # Track stats - make sure to get GO aspect from ont var
            if relation not in DISTINCT_EXTENSIONS:
                DISTINCT_EXTENSIONS[relation] = {}
            if term_prefix not in DISTINCT_EXTENSIONS[relation]:
                # should be checking for duplicate rows
                DISTINCT_EXTENSIONS[relation][term_prefix] = []
            # else:
            #     DISTINCT_EXTENSIONS[relation][term_prefix] += 1
            # if row_cols not in DISTINCT_EXTENSIONS[relation][term_prefix]:
            DISTINCT_EXTENSIONS[relation][term_prefix].append(row_cols)
        # ordering of extension relations may be inconsistent so sort
        return sorted(ext_list, key=str.lower)

    # Do we need this? Does this work? WE TOTALLY NEED THIS!!! At least maybe as an entry point to find what bucket an
    # annotation falls in.
    # HANDLE ASSOCIATION OBJECT_EXTENSIONS STRUCTURE
    def annot_following_rules(self, annot, aspect, term):
        ext_list = []
        #TODO for a in annot.split("|"):
        # extensions = annot.split(",")
        # annot["object_extensions"]
        # ext_list = self.extensions_list(extensions)
        ext_list = self.extensions_list(annot)
        # Standardize key - ex:
        #   ["part_of(GO:aspect),part_of(UBERON:)"]
        return self.following_rules(ext_list, aspect, term)


d = [
    "GeneDB_tsetse",
    "gonuts",
    "reactome",
    "isoform",
    "goa_pdb",
    "goa_uniprot_all.gaf",
]


if __name__ == "__main__":
    args = parser.parse_args()

    def get_filter_name(gpad_fname):
        if args.mod:
            filter_name = args.mod
        else:
            # Try matching filename to filter rule (e.g. wb.gpad to 'WB' to WBFilterRule)
            filter_name = os.path.basename(os.path.splitext(gpad_fname)[0]).upper()
        return filter_name

    bad_extensions = []

    # filenames = []
    # Use dict for holding different FilterRules for each file
    filenames = {}
    # data = []
    # fname = "/Users/ebertdu/go/go-pombase/gene_association.pombase-06-2018"
    # data = GafParser().parse(fname, skipheader=True)
    if args.filename is not None:
        # filenames.append(args.filename)
        filter_name = get_filter_name(args.filename)
        filter_rule = get_filter_rule(filter_name)
        filenames[args.filename] = filter_rule
        # data = GafParser().parse(args.filename, skipheader=True)
    elif args.dir is not None:
        for fname in os.listdir(args.dir):
            # print("Loading file:", fname)
            nono_in_fname = False
            for nono in d:
                if nono in fname:
                    nono_in_fname = True
            if fname.endswith(".tsv") or nono_in_fname:
                continue
            # filenames.append(args.dir + fname)
            filter_name = get_filter_name(fname)
            filenames[args.dir + fname] = get_filter_rule(filter_name)
            # data = data + GafParser().parse(fname, skipheader=True)

    # all_dict = {}
    extensions_mapper = ExtensionsMapper(constraints_yaml=args.constraints_yaml)
    gpad_parser = GpadParser()
    print("Creating extension dictionary...")
    ext_dict = {}
    ext_dict['F'] = {}
    ext_dict['P'] = {}
    ext_dict['C'] = {}
    for fname in filenames:
        with open(fname) as f:
            data = []
            print("Loading file:", fname)
            for l in f.readlines():
                if not l.startswith("!"):
                    parts = l.split("\t")
                    # if parts[15] != "" and parts[6] in acceptable_evidence_codes:
                    data.append(parts)
            print("# of GPAD lines in file:", len(data))
            data = filter_has_extension(data)
            print("# of GPAD lines having extensions:", len(data))
            # filter_rule = get_filter_rule(args.mod)
            filter_rule = filenames[fname]
            assoc_filter = AssocFilter(filter_rule)
            data = filter_rule_validate_lines(data, assoc_filter)
            print("Total GPAD count after applying {}: {}".format(filter_rule.__class__.__name__, len(data)))

            for g in data:
                go_term = g[3]
                aspect = extensions_mapper.go_aspector.go_aspect(go_term)
                split_line = SplitLine(line="\t".join(g), values=g, taxon="")  # Needed for ontobio error handling
                ontobio_extensions = gpad_parser._parse_full_extension_expression(g[10], split_line)
                ontobio_extensions = extensions_mapper.dedupe_extensions(ontobio_extensions)
                # ontobio_pattern = {
                #     'union_of': [
                #         {
                #             'intersection_of': [
                #                 {'property': 'part_of', 'filler': 'CL:0000678'},
                #                 {'property': 'part_of', 'filler': 'EMAPA:16525'}
                #             ]
                #         },
                #         {
                #               'intersection_of': [
                #                   {'property': 'part_of', 'filler': 'CL:0000678'},
                #                   {'property': 'part_of', 'filler': 'EMAPA:16525'}
                #               ]
                #         }
                #     ]
                # }
                for onto_ext in ontobio_extensions:
                    ext_list = extensions_mapper.extensions_list(onto_ext['intersection_of'], g)
                    check_ext_result = extensions_mapper.following_rules(ext_list, aspect, go_term)
                    if not check_ext_result.is_valid:
                        ext_key = ",".join(ext_list)
                        offending_extension = check_ext_result.offending_extension
                        if offending_extension is None:
                            offending_extension = check_ext_result.reason
                        bad_extensions.append([":".join(g[0:2]), go_term, GO_ONTOLOGY.label(go_term), offending_extension, g[10]])
                        if ext_key not in ext_dict[aspect]:
                            ext_dict[aspect][ext_key] = [g]
                        elif g not in ext_dict[aspect][ext_key]:
                            ext_dict[aspect][ext_key].append(g)

            for aspect in ['F','P','C']:
                max_count = 0
                top_k = None
                example_v = None
                for k, v in ext_dict[aspect].items():
                    if len(v) > max_count:
                        max_count = len(v)
                        # print(max_count)
                        top_k = k
                        example_v = v[0]

    def assigner_count(annots, assign):
        count = 0
        for a in annots:
            if a[9] == assign:
                count += 1
        return count

    cols = ['Aspect', 'Total count', 'Extension']
    all_assigners = []
    for aspect in ['F', 'P', 'C']:
        for k, v in ext_dict[aspect].items():
            for a in v:
                if a[9] not in all_assigners:
                    all_assigners.append(a[9])
                    cols.append(a[9])

    def parse_pattern_sourcefile(sourcefile):
        parsed_patterns = []
        with open(sourcefile) as sf:
            for pl in sf.readlines():
                pl = pl.rstrip()
                parsed_patterns.append(pl)
        return parsed_patterns

    class GpadWriter:
        def __init__(self, gpad_file):
            self.gpad_file = gpad_file
            self.writer = csv.writer(pattern_gpad, delimiter="\t")

        def writerow(self, row):
            self.writer.writerow(row)

    class WriterCollection:
        def __init__(self):
            self.writers = {}
            self.rows = {}

        def set_writer(self, writer, writer_name):
            self.writers[writer_name] = writer
            self.rows[writer_name] = []


    patterns = []
    writers = WriterCollection()
    if args.pattern or args.pattern_sourcefile:
        if args.pattern_sourcefile:
            patterns = parse_pattern_sourcefile(args.pattern_sourcefile)
            # writers = {}
            for patt in patterns:
                pattern_outfile = "{}.gpad".format(patt)
                pattern_gpad = open(pattern_outfile, "w+")
                pattern_gpad_writer = GpadWriter(pattern_gpad)
                writers.set_writer(pattern_gpad_writer, patt)
        else:
            patterns.append(args.pattern)
            pattern_outfile = "{}.gpad".format(args.pattern)
            if args.pattern_outfile:
                pattern_outfile = args.pattern_outfile
            pattern_gpad = open(pattern_outfile, "w+")
            pattern_gpad_writer = GpadWriter(pattern_gpad)
            writers.set_writer(pattern_gpad_writer, args.pattern)
    out_file = "all.tsv"
    if args.out_file:
        out_file = args.out_file
    with open(out_file, 'w') as f:
        writer = csv.writer(f, delimiter='\t')
        writer.writerow(cols)
        for aspect in ['F', 'P', 'C']:
            max_count = 0
            top_k = None
            example_v = None
            for k, v in ext_dict[aspect].items():
                row_to_write = [aspect, len(v), k]
                if len(patterns) > 0:
                    for patt in patterns:
                        # if k == patt:
                        if patt in k:
                            for a in v:
                                writers.rows[patt].append(a[0:len(a)-1])
                                # writers.writers[patt].writerow(a[0:len(a)-1])
                for assigner in all_assigners:
                    a_count = assigner_count(v, assigner)
                    row_to_write.append(a_count)
                writer.writerow(row_to_write)
    for writer_name in writers.writers:
        for row in writers.rows[writer_name]:
            writers.writers[writer_name].writerow(row)
        gpad_file = writers.writers[writer_name].gpad_file
        print("Offending GPAD lines written to:", gpad_file.name)
        gpad_file.close()

    wanted_gafs = []
    leftovers_out_file = "leftovers.gpad"
    if args.leftovers_out_file:
        leftovers_out_file = args.leftovers_out_file
    with open(leftovers_out_file, 'w') as wf:
        for k in ext_dict['P'].keys():
            if 'has_regulation_target' in k:
                wanted_gafs = wanted_gafs + ext_dict['P'][k]
                for g in ext_dict['P'][k]:
                    wf.write("\t".join(g))

    if args.extensions_list:
        with open("distinct_extensions.txt", "w+") as de_f:
            de_writer = csv.writer(de_f, delimiter="\t")
            de_writer.writerow(["pattern", "namespaces", "usage count", "line count", "RO code"])
            for rel in DISTINCT_EXTENSIONS:
                distinct_lines = []
                usage_count = 0
                ro_rel = extensions_mapper.translate_relation_to_ro(rel)
                for prefix in DISTINCT_EXTENSIONS[rel]:
                    usage_count += len(DISTINCT_EXTENSIONS[rel][prefix])
                    # distinct_lines = []
                    for line in DISTINCT_EXTENSIONS[rel][prefix]:
                        if line not in distinct_lines:
                            distinct_lines.append(line)
                    # line_count = len(distinct_lines)
                    # de_writer.writerow(["{}({})".format(rel, prefix), len(DISTINCT_EXTENSIONS[rel][prefix]), line_count])
                line_count = len(distinct_lines)
                de_writer.writerow([rel, ",".join(DISTINCT_EXTENSIONS[rel].keys()), usage_count, line_count, ro_rel])

    bad_extensions_fname = date_fname("bad_extensions.tsv")
    with open(bad_extensions_fname, "w+") as bef:
        be_writer = csv.writer(bef, delimiter="\t")
        be_writer.writerow(["DB Object", "Primary term", "Label", "Offending extension", "Full extensions"])
        for be in bad_extensions:
            be_writer.writerow(be)


import click
import yaml
import os
import glob

from dataclasses import dataclass

import yamldown

from typing import Dict, Set

"""
This module is for interacting with the different types of metadata that live
in the geneontology/go-site/metadata directory
"""

###############################################

class BiDiMultiMap(dict):

    def __init__(self, *args):
        super(BiDiMultiMap, self).__init__(*args)
        self.__reverse = dict()

    def __setitem__(self, key, val):

        # Turn everything into a set
        inval = None
        if isinstance(val, set):
            inval = val
        elif isinstance(val, iter):
            inval = set(val)
        elif val == None:
            inval = set()
        else:
            inval = set([val])

        olds = self.get(key, [])
        super(BiDiMultiMap, self).__setitem__(key, inval)
        # Remove old value mappings from the key
        for old in olds:
            if old in self.__reverse:
                del self.__reverse[old]

        # Add the items
        for v in inval:
            self.__reverse[v] = key

    def __delitem__(self, key):
        vs = self[key]
        super(BiDiMultiMap, self).__delitem__(key)
        for v in vs:
            del self.__reverse[v]


    def reverse(self, element):
        """
        Returns the key that has element in the set for the mapping of `key`, or
        None, of there is no such element.

        Example:
        Given "FB" --> set("fb", "FlyBase")

        reverse("fb") --> "FB"
        reverse("FlyBase") --> "FB"
        reverse("Foo") --> None
        """
        return self.__reverse.get(element, None)

# d = BiDiMultiMap()
#
# d["FB"] = set(["fb", "FlyBase"])
# d.reverse("fb") # returns "FB"
# d.reverse("FlyBase") # returns "FB"
# d.reverse("mgi:mgi") # returns None

###############################################
#
# @dataclass
# class YamldownMetadata:
#     id: str
#     layout: str
#     _meta: dict
#     _text: str


def yamldown_lookup(yamldown_dir):
    d = {
        metayaml_id(meta_path): get_yamldown_metadata(yamldown_dir, metayaml_id(meta_path))
        for meta_path in glob.glob("{}/*.md".format(os.path.join(yamldown_dir))) if metayaml_id(meta_path) not in ["ABOUT", "README", "SOP", "README-editors"]
    }

    return d

def dataset_metadata_file(metadata, group, empty_ok=False) -> Dict:
    metadata_yaml = os.path.join(metadata, "datasets", "{}.yaml".format(group))
    try:
        with open(metadata_yaml, "r") as group_data:
            click.echo("Found {group} metadata at {path}".format(group=group, path=metadata_yaml))
            return yaml.load(group_data, Loader=yaml.FullLoader)
    except Exception as e:
        if not empty_ok:
            raise click.ClickException("Could not find or read {}: {}".format(metadata_yaml, str(e)))
        else:
            return None

def extensions_constraints_file(metadata):
    constraints_path = os.path.join(metadata, "extensions-constraints.yaml")
    try:
        with open(constraints_path, "r") as constraints_file:
            click.echo("Found groups at {path}".format(path=constraints_path))
            constraints_list = yaml.load(constraints_file, Loader=yaml.FullLoader)
    except Exception as e:
        raise click.ClickException("Could not find or read {}: {}".format(constraints_path, str(e)))

    return constraints_list


def gorule_title(metadata, rule_id) -> str:
    return gorule_metadata(metadata, rule_id)["title"]

def gorule_metadata(metadata, rule_id) -> dict:
    gorule_yamldown = os.path.join(metadata, "rules")
    return get_yamldown_metadata(gorule_yamldown, rule_id)

def parse_goref_metadata(metadata, goref_id) -> dict:
    goref_yamldown = os.path.join(metadata, "gorefs")
    return get_yamldown_metadata(goref_yamldown, goref_id)

def get_yamldown_metadata(yamldown_dir, meta_id) -> dict:
    yamldown_md_path = os.path.join(yamldown_dir, "{}.md".format(meta_id))
    try:
        with open(yamldown_md_path, "r") as gorule_data:
            return yamldown.load(gorule_data)[0]
    except Exception as e:
        raise click.ClickException("Could not find or read {}: {}".format(yamldown_md_path))

def metayaml_id(rule_path) -> str:
    return os.path.splitext(os.path.basename(rule_path))[0]

def source_path(dataset_metadata, target_dir, group):
    extension = dataset_metadata["type"]
    if dataset_metadata.get("compression", False):
        extension = "{ext}.gz".format(ext=extension)

    path = os.path.join(target_dir, "groups", group, "{name}-src.{ext}".format(name=dataset_metadata["dataset"], ext=extension))
    return path

def database_entities(metadata):
    dbxrefs_path = os.path.join(os.path.abspath(metadata), "db-xrefs.yaml")
    try:
        with open(dbxrefs_path, "r") as db_xrefs_file:
            click.echo("Found db-xrefs at {path}".format(path=dbxrefs_path))
            dbxrefs = yaml.load(db_xrefs_file, Loader=yaml.FullLoader)
    except Exception as e:
        raise click.ClickException("Could not find or read {}: {}".format(dbxrefs_path, str(e)))

    d = BiDiMultiMap()
    for entity in dbxrefs:
        d[entity["database"]] = set(entity.get("synonyms", []))

    return d

def groups(metadata) -> Set[str]:
    groups_path = os.path.join(os.path.abspath(metadata), "groups.yaml")
    try:
        with open(groups_path, "r") as groups_file:
            click.echo("Found groups at {path}".format(path=groups_path))
            groups_list = yaml.load(groups_file, Loader=yaml.FullLoader)
    except Exception as e:
        raise click.ClickException("Could not find or read {}: {}".format(groups_path, str(e)))

    return set([group["shorthand"] for group in groups_list])



from ontobio.rdfgen.assoc_rdfgen import CamRdfTransform, TurtleRdfWriter, genid, prefix_context
from ontobio.rdfgen import relations
from ontobio.vocabulary.relations import OboRO, Evidence
from ontobio.vocabulary.upper import UpperLevel
from prefixcommons.curie_util import expand_uri, contract_uri
from rdflib.namespace import OWL, RDF
from rdflib import Literal
from rdflib.term import URIRef
from rdflib.namespace import Namespace
import rdflib
import datetime
import dateutil.parser
import os.path as path
import logging
from typing import List
from ontobio.rdfgen.gocamgen.triple_pattern_finder import TriplePattern, TriplePatternFinder
from ontobio.rdfgen.gocamgen.subgraphs import AnnotationSubgraph
from ontobio.rdfgen.gocamgen.collapsed_assoc import CollapsedAssociationSet, CollapsedAssociation, dedupe_extensions, CollapsedAssociationLine
from ontobio.rdfgen.gocamgen.utils import sort_terms_by_ontology_specificity, ShexHelper, GroupsHelper
from ontobio.rdfgen.gocamgen import errors
from ontobio.model.association import GoAssociation, ExtensionUnit, ConjunctiveSet, ymd_str
from ontobio.io.assocparser import AssocParserConfig


# logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)
logger.setLevel("INFO")

ro = OboRO()
evt = Evidence()
upt = UpperLevel()
LEGO = Namespace("http://geneontology.org/lego/")
LAYOUT = Namespace("http://geneontology.org/lego/hint/layout/")
PAV = Namespace('http://purl.org/pav/')
DC = Namespace("http://purl.org/dc/elements/1.1/")
RDFS = Namespace("http://www.w3.org/2000/01/rdf-schema#")
GOREL = Namespace("http://purl.obolibrary.org/obo/GOREL_")
NCBITAXON = Namespace("http://purl.obolibrary.org/obo/NCBITaxon_")
BIOLINK = Namespace("https://w3id.org/biolink/vocab/")

# Stealing a lot of code for this from ontobio.rdfgen:
# https://github.com/biolink/ontobio


def expand_uri_wrapper(id):
    c = prefix_context
    c['GOREL'] = "http://purl.obolibrary.org/obo/GOREL_"
    uri = expand_uri(id, cmaps=[c])
    return uri

def contract_uri_wrapper(id):
    uri = contract_uri(id, cmaps=[prefix_context])
    return uri

HAS_SUPPORTING_REFERENCE = URIRef(expand_uri_wrapper("dc:source"))
ENABLED_BY = URIRef(expand_uri_wrapper(ro.enabled_by))
ENABLES = URIRef(expand_uri_wrapper(ro.enables))
INVOLVED_IN = URIRef(expand_uri_wrapper(ro.involved_in))
PART_OF = URIRef(expand_uri_wrapper(ro.part_of))
OCCURS_IN = URIRef(expand_uri_wrapper(ro.occurs_in))
COLOCALIZES_WITH = URIRef(expand_uri_wrapper(ro.colocalizes_with))
CONTRIBUTES_TO = URIRef(expand_uri_wrapper("RO:0002326"))
MOLECULAR_FUNCTION = URIRef(expand_uri_wrapper(upt.molecular_function))
REGULATES = URIRef(expand_uri_wrapper("RO:0002211"))
LOCATED_IN = URIRef(expand_uri_wrapper("RO:0001025"))

# RO_ONTOLOGY = OntologyFactory().create("http://purl.obolibrary.org/obo/ro.owl")  # Need propertyChainAxioms to parse (https://github.com/biolink/ontobio/issues/312)
INPUT_RELATIONS = {
    #TODO Create rule for deciding (MOD-specific?) whether to convert has_direct_input to has input
    # "has_direct_input": "RO:0002400",
    "has_direct_input": "RO:0002233",
    "has input": "RO:0002233",
    "has_input": "RO:0002233",
    "occurs_in": "BFO:0000066",
    "has_agent": "RO:0002218"
}
ACTS_UPSTREAM_OF_RELATIONS = {
    "acts_upstream_of": "RO:0002263",
    "acts_upstream_of_or_within": "RO:0002264",
    "acts upstream of or within, positive effect": "RO:0004032",
    "acts upstream of or within, negative effect": "RO:0004033",
    "acts_upstream_of_positive_effect": "RO:0004034",
    "acts_upstream_of_negative_effect": "RO:0004035",
}
HAS_REGULATION_TARGET_RELATIONS = {
    # WB:WBGene00013591 involved_in GO:0042594
    "has_regulation_target": ""
}
REGULATES_CHAIN_RELATIONS = [
    "regulates_o_occurs_in",
    "regulates_o_acts_on_population_of"
]


SHEX_HELPER = ShexHelper()
GROUPS_HELPER = GroupsHelper()


def has_regulation_target_bucket(ontology, term):
    ancestors = ontology.ancestors(term, reflexive=True)
    buckets = []
    if "GO:0065009" in ancestors:
        buckets.append("a")
    if "GO:0010468" in ancestors:
        buckets.append("b")
    if "GO:0002092" in ancestors or "GO:0042176" in ancestors:
        buckets.append("c")
    if "GO:0019538" in ancestors or "GO:0032880" in ancestors:
        buckets.append("d")
    return buckets


now = datetime.datetime.now()


class Annoton():
    def __init__(self, subject_id, assocs, connections=None):
        self.enabled_by = subject_id
        self.annotations = assocs
        self.connections = connections
        self.individuals = {}


class GoCamEvidence:
    DEFAULT_CONTRIBUTOR = "http://orcid.org/0000-0002-6659-0416"
    DEFAULT_PROVIDED_BY = "http://geneontology.org"  # GO in groups.yaml

    def __init__(self, code, references, contributors=[], date="", comment="", with_from=None, provided_by=[]):
        self.evidence_code = code
        self.references = references
        self.date = date
        self.contributors = contributors
        self.provided_by = provided_by
        self.comment = comment
        self.with_from = with_from
        self.id = None

    @staticmethod
    def create_from_annotation(annot: CollapsedAssociationLine):
        source_line = annot.source_line.rstrip().replace("\t", " ")
        contributors = []
        if "contributor-id" in annot.annotation_properties:
            contributors = annot.annotation_properties["contributor-id"]
        if len(contributors) == 0:
            contributors = [GoCamEvidence.DEFAULT_CONTRIBUTOR]
        provided_by = GROUPS_HELPER.lookup_group_id(annot.assigned_by)  # Needs to be resolved to URI, e.g. ZFIN->http://zfin.org

        return GoCamEvidence(annot.evidence_code, annot.references,
                           contributors=contributors,
                           # "{date}T{time}".format(date=ymd_str(date), time=date.time)
                           date=ymd_str(annot.date, separator="-"),
                           provided_by=provided_by,
                           comment=source_line)

    @staticmethod
    def create_from_collapsed_association(collapsed_association: CollapsedAssociation):
        evidences = []
        for line in collapsed_association:
            evidence = GoCamEvidence.create_from_annotation(line)
            if line.with_from:
                evidence.with_from = ",".join(line.with_from)
            evidences.append(evidence)
        return evidences

    @staticmethod
    def max_date(evidences: List):
        all_dates = [e.date for e in evidences]
        mdate = sorted(all_dates, key=lambda x: dateutil.parser.parse(x), reverse=True)[0]
        return mdate


class GoCamModel():
    # TODO: Not using anymore maybe get rid of?
    relations_dict = {
        "has_direct_input": "RO:0002400",
        "has input": "RO:0002233",
        "has_regulation_target": "RO:0002211",  # regulates
        "regulates_activity_of": "RO:0002578",  # directly regulates
        "with_support_from": "RO:0002233",  # has input
        "directly_regulates": "RO:0002578",
        "directly_positively_regulates": "RO:0002629",
        "directly_negatively_regulates": "RO:0002630",
        "colocalizes_with": "RO:0002325",
        "contributes_to": "RO:0002326",
        "part_of": "BFO:0000050",
        "acts_upstream_of": "RO:0002263",
        "acts_upstream_of_negative_effect": "RO:0004035",
        "acts_upstream_of_or_within": "RO:0002264",
        "acts_upstream_of_positive_effect": "RO:0004034",
        "acts upstream of, negative effect": "RO:0004035",
        "acts_upstream_of_or_within_negative_effect": "RO:0004033",
        "acts_upstream_of_or_within_positive_effect": "RO:0004032",
        "located_in": "RO:0001025",
    }

    def __init__(self, modeltitle, connection_relations=None, store=None, model_id=None):
        self.modeltitle = modeltitle
        cam_writer = CamTurtleRdfWriter(self.modeltitle, store=store, model_id=model_id)
        self.writer = AnnotonCamRdfTransform(cam_writer)
        self.classes = []
        self.individuals = {}   # Maintain entity-to-IRI dictionary. Prevents dup individuals but we may want dups?
        # TODO: Refactor to make graph more prominent
        self.graph = self.writer.writer.graph
        if connection_relations is None:
            self.connection_relations = GoCamModel.relations_dict
        else:
            self.connection_relations = connection_relations
        self.declare_properties()

    def write(self, filename, format='ttl'):
        if path.splitext(filename)[1] != ".ttl":
            filename += ".ttl"
        with open(filename, 'wb') as f:
            self.writer.writer.serialize(destination=f, format=format)

    def declare_contributor(self, contributor: str):
        self.graph.add((self.writer.writer.base, DC.contributor, Literal(contributor)))

    def declare_provided_by(self, provided_by: str):
        self.graph.add((self.writer.writer.base, PAV.providedBy, Literal(provided_by)))

    def declare_properties(self):
        # AnnotionProperty
        self.writer.emit_type(URIRef("http://geneontology.org/lego/evidence"), OWL.AnnotationProperty)
        self.writer.emit_type(URIRef("http://geneontology.org/lego/hint/layout/x"), OWL.AnnotationProperty)
        self.writer.emit_type(URIRef("http://geneontology.org/lego/hint/layout/y"), OWL.AnnotationProperty)
        self.writer.emit_type(URIRef("http://purl.org/pav/providedBy"), OWL.AnnotationProperty)
        self.writer.emit_type(URIRef("http://purl.org/dc/elements/1.1/contributor"), OWL.AnnotationProperty)
        self.writer.emit_type(URIRef("http://purl.org/dc/elements/1.1/date"), OWL.AnnotationProperty)
        self.writer.emit_type(URIRef("http://purl.org/dc/elements/1.1/source"), OWL.AnnotationProperty)

    def declare_class(self, class_id):
        if class_id not in self.classes:
            self.writer.emit_type(URIRef(expand_uri_wrapper(class_id)), OWL.Class)
            self.classes.append(class_id)

    def declare_individual(self, entity_id, evidences: List[GoCamEvidence] = None, negated=False):
        entity = genid(base=self.writer.writer.base + '/')
        # TODO: Make this add_to_graph
        if negated:
            self.writer.emit_not(entity, self.writer.uri(entity_id))
        else:
            self.writer.emit_type(entity, self.writer.uri(entity_id))
        self.writer.emit_type(entity, OWL.NamedIndividual)
        if evidences:
            # Emit max date all contributors
            max_date = GoCamEvidence.max_date(evidences)
            self.writer.emit(entity, DC.date, Literal(max_date))
            all_contributors = set()
            for e in evidences:
                [all_contributors.add(c) for c in e.contributors]
            for c in all_contributors:
                self.writer.emit(entity, DC.contributor, Literal(c))
        self.individuals[entity_id] = entity
        return entity

    def add_axiom(self, statement, evidence=None):
        (source_id, property_id, target_id) = statement
        stmt_id = self.find_bnode(statement)
        if stmt_id is None:
            stmt_id = self.writer.blanknode()
            self.writer.emit_type(stmt_id, OWL.Axiom)
        self.writer.emit(stmt_id, OWL.annotatedSource, source_id)
        self.writer.emit(stmt_id, OWL.annotatedProperty, property_id)
        self.writer.emit(stmt_id, OWL.annotatedTarget, target_id)
        self.writer.emit_type(property_id, OWL.ObjectProperty)

        if evidence:
            self.add_evidence(stmt_id, evidence.evidence_code, evidence.references)

        return stmt_id

    def create_axiom(self, subject_id, relation_uri, object_id):
        # If not URIRef, will be class IDs
        subject_uri = subject_id if isinstance(subject_id, URIRef) else self.declare_individual(subject_id)
        object_uri = object_id if isinstance(object_id, URIRef) else self.declare_individual(object_id)
        axiom_id = self.add_axiom(self.writer.emit(subject_uri, relation_uri, object_uri))
        return axiom_id

    # TODO: Explicitly type subject, object parameters. Are they Class ID URIs or instance URIs?
    # def find_or_create_axiom_by_class_id_uri
    # def find_or_create_axiom_by_instance_uri
    def find_or_create_axiom(self, subject_id : str, relation_uri : URIRef, object_id : str, annoton=None,
                             exact_length=False):
        # Maybe overkill but gonna try using find_pattern_recursive to find only one triple
        # TODO: Replace the TriplePattern stuff w/ SPARQL but need 'exact_length' to work
        pattern = TriplePattern([(subject_id, relation_uri, object_id)])
        found_triples = TriplePatternFinder().find_pattern_recursive(self, pattern, exact_length=True)
        # found_triples = self.triples_by_ids(subject_id, relation_uri, object_id)
        if len(found_triples) > 0:
            # Gonna be a list of "triple-chains", itself a list of triples, and each triple is sort like a 3-index list.
            # So we just want the first triple from the first chain:
            found_triple = found_triples[0][0]
            subject_uri = found_triple[0]
            object_uri = found_triple[2]
            axiom_id = self.find_bnode(found_triple)
        else:
            # subject_uri = self.declare_individual(subject_id)
            subject_uri = subject_id if isinstance(subject_id, URIRef) else self.declare_individual(subject_id)
            object_uri = object_id if isinstance(object_id, URIRef) else self.declare_individual(object_id)
            # TODO Can emit() be changed to emit_axiom()?
            axiom_id = self.add_axiom(self.writer.emit(subject_uri, relation_uri, object_uri))
        if annoton and relation_uri == ENABLED_BY:
            annoton.individuals[subject_id] = subject_uri
            annoton.individuals[object_id] = object_uri
        return axiom_id

    def add_evidence(self, axiom, evidence: GoCamEvidence, emit_date=True):
        # Try finding existing evidence object containing same type and references
        # ev_id = self.writer.find_or_create_evidence_id(ev)
        ev_id = self.writer.create_evidence(evidence)
        self.writer.emit(axiom, URIRef("http://geneontology.org/lego/evidence"), ev_id)
        ### Emit ev fields to axiom here TODO: Couple evidence and axiom emitting together
        self.writer.emit(axiom, RDFS.comment, Literal(evidence.comment))
        self.writer.emit(axiom, PAV.providedBy, Literal(evidence.provided_by))
        for c in evidence.contributors:
            self.writer.emit(axiom, DC.contributor, Literal(c))
        if emit_date:
            self.writer.emit(axiom, DC.date, Literal(evidence.date))

    def add_evidences(self, axiom, evidences: List[GoCamEvidence]):
        for e in evidences:
            self.add_evidence(axiom, e, emit_date=False)
        max_date = GoCamEvidence.max_date(evidences)
        self.writer.emit(axiom, DC.date, Literal(max_date))

    def add_connection(self, gene_connection, source_annoton):
        # Switching from reusing existing activity node from annoton to creating new one for each connection - Maybe SPARQL first to check if annoton activity already used for connection?
        # Check annoton for existing activity.
        # if gene_connection.object_id in source_annoton.individuals:
        #     # If exists and activity has connection relation,
        #     # Look for two triples: (gene_connection.object_id, ENABLED_BY, source_annoton.enabled_by) and (gene_connection.object_id, connection_relations, anything)
        # Annot MF should be declared by now - don't declare object_id if object_id == annot MF?
        if gene_connection.gp_b not in self.individuals:
            return
        source_id = None
        uri_list = self.uri_list_for_individual(gene_connection.object_id)
        for u in uri_list:
            if gene_connection.relation in self.connection_relations:
                rel = URIRef(expand_uri_wrapper(self.connection_relations[gene_connection.relation]))
                # Annot MF should be declared by now - don't declare object_id if object_id == annot MF?
                try:
                    annot_mf = source_annoton.molecular_function["object"]["id"]
                except:
                    annot_mf = ""
                if (u,rel,None) in self.writer.writer.graph and gene_connection.object_id != annot_mf:
                    source_id = self.declare_individual(gene_connection.object_id)
                    source_annoton.individuals[gene_connection.object_id] = source_id
                    break

        if source_id is None:
            try:
                source_id = source_annoton.individuals[gene_connection.object_id]
            except KeyError:
                source_id = self.declare_individual(gene_connection.object_id)
                source_annoton.individuals[gene_connection.object_id] = source_id
        # Add enabled by stmt for object_id - this is essentially adding another annoton connecting gene-to-extension/with-MF to the model
        self.writer.emit(source_id, ENABLED_BY, source_annoton.individuals[source_annoton.enabled_by])
        self.writer.emit_axiom(source_id, ENABLED_BY, source_annoton.individuals[source_annoton.enabled_by])
        property_id = URIRef(expand_uri_wrapper(self.connection_relations[gene_connection.relation]))
        target_id = self.individuals[gene_connection.gp_b]
        # Annotate source MF GO term NamedIndividual with relation code-target MF term URI
        self.writer.emit(source_id, property_id, target_id)
        # Add axiom (Source=MF term URI, Property=relation code, Target=MF term URI)
        self.writer.emit_axiom(source_id, property_id, target_id)

    def uri_list_for_individual(self, individual):
        uri_list = []
        graph = self.writer.writer.graph
        for t in graph.triples((None,None,self.writer.uri(individual))):
            uri_list.append(t[0])
        return uri_list

    def triples_by_ids(self, subject, relation_uri, object_id):
        graph = self.writer.writer.graph

        triples = []
        if isinstance(subject, URIRef) or subject is None:
            subjects = [subject]
        else:
            subjects = self.uri_list_for_individual(subject)
        if isinstance(object_id, URIRef) or object_id is None:
            objects = [object_id]
        else:
            objects = self.uri_list_for_individual(object_id)
        for object_uri in objects:
            for subject_uri in subjects:
                # if (subject_uri, relation_uri, object_uri) in graph:
                #     triples.append((subject_uri, relation_uri, object_uri))
                for t in graph.triples((subject_uri, relation_uri, object_uri)):
                    triples.append(t)
        return triples

    def individual_label_for_uri(self, uri):
        ind_list = []
        graph = self.writer.writer.graph
        for t in graph.triples((uri, RDF.type, None)):
            if t[2] != OWL.NamedIndividual: # We know OWL.NamedIndividual triple does't contain the label so don't return it
                ind_list.append(t[2])
        return ind_list

    def class_for_uri(self, uri):
        try:
            class_curie = contract_uri_wrapper(self.individual_label_for_uri(uri)[0])[0]
            return class_curie
        except:
            return None

    def axioms_for_source(self, source, property_uri=None):
        if property_uri is None:
            property_uri = OWL.annotatedSource
        axiom_list = []
        graph = self.writer.writer.graph
        for uri in self.uri_list_for_individual(source):
            for t in graph.triples((None, property_uri, uri)):
                axiom_list.append(t[0])
        return axiom_list

    def find_bnode(self, triple):
        (subject,predicate,object_id) = triple
        s_triples = self.writer.writer.graph.triples((None, OWL.annotatedSource, subject))
        s_bnodes = [s for s,p,o in s_triples]
        p_triples = self.writer.writer.graph.triples((None, OWL.annotatedProperty, predicate))
        p_bnodes = [s for s,p,o in p_triples]
        o_triples = self.writer.writer.graph.triples((None, OWL.annotatedTarget, object_id))
        o_bnodes = [s for s,p,o in o_triples]
        bnodes = set(s_bnodes) & set(p_bnodes) & set(o_bnodes)
        if len(bnodes) > 0:
            return list(bnodes)[0]

    def triples_involving_individual(self, ind_id, relation=None):
        # "involving" meaning individual (URI) is either subject or object
        graph = self.writer.writer.graph
        found_triples = list(graph.triples((ind_id, relation, None)))
        for t in graph.triples((None, relation, ind_id)):
            if t not in found_triples:
                found_triples.append(t)
        return found_triples


def relation_equals(rel_a, rel_b):
    if ":" not in str(rel_a):
        rel_a = contract_uri_wrapper(relations.lookup_label(rel_a))
    if ":" not in str(rel_b):
        rel_b = contract_uri_wrapper(relations.lookup_label(rel_b))[0]

    # Get to curie str to compare
    result = str(rel_a) == str(rel_b)
    return result


class AssocGoCamModel(GoCamModel):
    ENABLES_O_RELATION_LOOKUP = {}

    def __init__(self, modeltitle, assocs: List[GoAssociation], config: AssocParserConfig=None, connection_relations=None, store=None, gpi_entities=None, model_id=None):
        GoCamModel.__init__(self, modeltitle, connection_relations, store, model_id=model_id)
        self.ontology = config.ontology
        self.associations = CollapsedAssociationSet(ontology=self.ontology, gpi_entities=gpi_entities)
        self.associations.collapse_annotations(assocs)
        self.go_aspector = None  # TODO: Can I always grab aspect from ontology term DS
        self.default_contributor = "http://orcid.org/0000-0002-6659-0416"
        self.contributors = set()
        self.provided_bys = set()
        self.graph.bind("GOREL", GOREL)  # Because GOREL isn't in context.jsonld's
        self.gpi_entities = gpi_entities
        ncbi_taxon = self.taxon_id_from_entity(str(assocs[0].subject.id))
        # Emit model-level in_taxon triple from ncbi_taxon
        if ncbi_taxon:
            # <https://w3id.org/biolink/vocab/in_taxon> <http://purl.obolibrary.org/obo/NCBITaxon_10090>
            self.graph.add((self.writer.writer.base, BIOLINK.in_taxon, URIRef(expand_uri_wrapper(ncbi_taxon))))
        self.graph.add((self.writer.writer.base, DC.title, Literal(modeltitle)))

    def taxon_id_from_entity(self, entity_id: str):
        if entity_id in self.gpi_entities:
            return self.gpi_entities[entity_id]["taxon"]["id"]
        return None

    def translate(self):

        for a in self.associations:

            term = str(a.object_id())

            # Add evidences tied to axiom_ids
            evidences = GoCamEvidence.create_from_collapsed_association(a)
            # Record all contributors at model level
            for e in evidences:
                for c in e.contributors:
                    self.contributors.add(c)
                self.provided_bys.add(e.provided_by)

            annotation_extensions = a.annot_extensions()

            # Translate extension - maybe add function argument for custom translations?
            if not annotation_extensions:
                annot_subgraph = self.translate_primary_annotation(a)
                # For annots w/o extensions, this is where we write subgraph to model
                annot_subgraph.write_to_model(self, evidences)
            else:
                aspect = self.go_aspector.go_aspect(term)

                # TODO: Handle deduping in collapsed_assoc - need access to extensions_mapper.dedupe_extensions
                annotation_extensions = dedupe_extensions(annotation_extensions)

                # Split on those multiple occurs_in(same NS) extensions
                # TODO: Cleanup/refactor this splitting into separate method
                extension_sets_to_remove = []
                for uo in annotation_extensions:
                    # Grab occurs_in's
                    # Make a new uo if situation found
                    occurs_in_exts: List[ExtensionUnit] = [ext for ext in uo.elements if relation_equals(ext.relation, "occurs_in")]
                    # onto_grouping = {
                    #     "CL": [{}, {}],
                    #     "EMAPA": [{}]
                    # }
                    onto_grouping = {}
                    for ext in occurs_in_exts:
                        ont_prefix = ext.term.namespace
                        if ont_prefix not in onto_grouping:
                            onto_grouping[ont_prefix] = []
                        onto_grouping[ont_prefix].append(ext)
                    for ont_prefix, exts in onto_grouping.items():
                        if len(exts) > 1:
                            if uo not in extension_sets_to_remove:
                                extension_sets_to_remove.append(uo)  # Remove original set when we're done splitting
                            for ext in exts:
                                # Create new 'intersection_of' list
                                new_exts_list = []
                                # Add ext to this new list if its prefix is not ont_prefix
                                for int_of_ext in uo.elements:
                                    if not relation_equals(int_of_ext.relation, "occurs_in") or int_of_ext.term.namespace != ont_prefix:
                                        # Add the extensions that don't currently concern us
                                        new_exts_list.append(int_of_ext)
                                # Then add occurs_in ext in current iteration
                                new_exts_list.append(ext)
                                annotation_extensions.append(ConjunctiveSet(elements=new_exts_list))
                # Remove original, un-split extension from list so it isn't translated
                [annotation_extensions.remove(ext_set) for ext_set in extension_sets_to_remove]

                for uo in annotation_extensions:
                    int_bits = []
                    for rel in uo.elements:
                        int_bits.append(str(rel))
                    ext_str = ",".join(int_bits)

                    annot_subgraph = self.translate_primary_annotation(a)

                    intersection_extensions: List[ExtensionUnit] = dedupe_extensions(uo.elements)
                    # is_cool = self.extensions_mapper.annot_following_rules(intersection_extensions, aspect, term)
                    is_cool = True  # Open the flood gates
                    if is_cool:
                        logger.debug("GOOD: {}".format(ext_str))
                        # Nesting repeated extension relations (i.e. occurs_in, part_of)
                        ext_rels_to_nest = ['occurs_in', 'part_of']  # Switch to turn on/off extension nesting
                        for ertn in ext_rels_to_nest:
                            nest_exts = [ext for ext in intersection_extensions if relation_equals(ext.relation, ertn)]
                            if len(nest_exts) > 1:
                                # Sort by specific term to general term
                                sorted_nest_ext_terms = sort_terms_by_ontology_specificity(
                                    [ne.term for ne in nest_exts])
                                # Translate
                                loc_subj_n = annot_subgraph.get_anchor()
                                for idx, ne_term in enumerate(sorted_nest_ext_terms):
                                    # location_relation could be part_of, occurs_in, or located_in
                                    # Figure out what types of classes these are
                                    # Use case here is matching to ShEx shape class
                                    subj_shape = SHEX_HELPER.shape_from_class(AnnotationSubgraph.node_class(loc_subj_n),
                                                                              self.go_aspector)
                                    loc_obj_n = annot_subgraph.add_instance_of_class(ne_term)
                                    obj_shape = SHEX_HELPER.shape_from_class(str(ne_term),
                                                                             self.go_aspector)
                                    # location_relation = "BFO:0000050"  # part_of
                                    location_relation = SHEX_HELPER.relation_lookup(subj_shape, obj_shape)
                                    if idx == 0 and ertn == "occurs_in":
                                        location_relation = "BFO:0000066"  # occurs_in - because MF -> @<AnatomicalEntity> OR @<CellularComponent>
                                    annot_subgraph.add_edge(loc_subj_n, location_relation, loc_obj_n)

                                    loc_subj_n = loc_obj_n  # For next iteration
                                # Remove from intersection_extensions because this is now already translated
                                [intersection_extensions.remove(ext) for ext in nest_exts]
                        for rel in intersection_extensions:
                            ext_relation = relations.lookup_uri(expand_uri_wrapper(str(rel.relation)), default="")
                            if ext_relation == "":
                                # AttributeError: 'NoneType' object has no attribute 'replace'
                                error_msg = "Relation '{}' not found in relations lookup. Skipping annotation translation.".format(str(rel.relation))
                                raise errors.GocamgenException(error_msg)
                            ext_target = str(rel.term)
                            if ext_relation not in list(INPUT_RELATIONS.keys()) + list(HAS_REGULATION_TARGET_RELATIONS.keys()):
                                # No RO term yet. Try looking up in RO
                                relation_term = self.translate_relation_to_ro(ext_relation)
                                if relation_term:
                                    # print("Ext relation {} auto-mapped to {} in {}".format(ext_relation, relation_term, a.subject_id()))
                                    INPUT_RELATIONS[ext_relation] = relation_term
                            if ext_relation in INPUT_RELATIONS:
                                ext_target_n = annot_subgraph.add_instance_of_class(ext_target)
                                # Need to find what mf we're talking about
                                anchor_n = annot_subgraph.get_anchor()
                                annot_subgraph.add_edge(anchor_n, INPUT_RELATIONS[ext_relation], ext_target_n)
                            # elif ext_relation in REGULATES_CHAIN_RELATIONS:
                            elif ext_relation.startswith("regulates_o_"):
                                # Get target MF from primary term (BP) e.g. GO:0007346 regulates some mitotic cell cycle
                                regulates_rel, regulated_term = self.get_rel_and_term_in_logical_definitions(term)
                                if regulates_rel:
                                    # Need to derive chained relation (e.g. "occurs_in") from this ext rel. Just replace("regulates_o_", "")?
                                    chained_rel_label = ext_relation.replace("regulates_o_", "")
                                    chained_rel = INPUT_RELATIONS.get(chained_rel_label)
                                    if chained_rel is None:
                                        chained_rel = self.translate_relation_to_ro(chained_rel_label)
                                    if chained_rel is None:
                                        # We have tried all we can for this poor thing. Skip translating this extension.
                                        continue
                                    regulated_term_n = annot_subgraph.add_instance_of_class(regulated_term)
                                    anchor_n = annot_subgraph.get_anchor()
                                    annot_subgraph.add_edge(anchor_n, regulates_rel, regulated_term_n)
                                    ext_target_n = annot_subgraph.add_instance_of_class(ext_target)
                                    annot_subgraph.add_edge(regulated_term_n, chained_rel, ext_target_n)
                                else:
                                    logger.warning("Couldn't get regulates relation from LD of: {}".format(term))
                            elif ext_relation in HAS_REGULATION_TARGET_RELATIONS:
                                if aspect == 'P':
                                    # For BP annotations, translate 'has regulation target' to 'has input'.
                                    ext_target_n = annot_subgraph.add_instance_of_class(ext_target)
                                    anchor_n = annot_subgraph.get_anchor()
                                    annot_subgraph.add_edge(anchor_n, INPUT_RELATIONS['has_input'], ext_target_n)
                                else:
                                    buckets = has_regulation_target_bucket(self.ontology, term)
                                    if len(buckets) > 0:
                                        bucket = buckets[0]  # Or express all buckets?
                                        # Four buckets
                                        if bucket in ["a", "d"]:
                                            regulates_rel, regulated_mf = self.get_rel_and_term_in_logical_definitions(term)
                                            if regulates_rel and regulated_mf:
                                                # [GP-A]<-enabled_by-[root MF]-regulates->[molecular function Z]-enabled_by->[GP-B]
                                                ext_target_n = annot_subgraph.add_instance_of_class(ext_target)
                                                regulated_mf_n = annot_subgraph.add_instance_of_class(regulated_mf)
                                                annot_subgraph.add_edge(regulated_mf_n, ro.enabled_by, ext_target_n)
                                                anchor_n = annot_subgraph.get_anchor()
                                                annot_subgraph.add_edge(anchor_n, regulates_rel, regulated_mf_n)
                                                # TODO: Suppress/delete (GP-A)<-enabled_by-(root MF)-part_of->(term) aka involved_in_translated
                                                # Remove (anchor_uri, None, term)
                                                # Is this anchor_uri always going to the root_mf?
                                                # Will the term individual be used for anything else?
                                                #   Other comma-delimited extensions on same annotation?
                                                # has_during? occurs_in?
                                                # WB:WBGene00001173 GO:0051343 ['has_regulation_target', 'occurs_in'] ['a']
                                                # WB:WBGene00006652 GO:0045944 ['has_regulation_target', 'occurs_in'] ['b']
                                                # WB:WBGene00003639 GO:0036003 ['happens_during', 'has_regulation_target'] ['b']
                                                # Other comma-delimited extensions (e.g. happens_during, has_input) need this triple?
                                                # WB:WBGene00002335 GO:1902685 ['has_regulation_target', 'occurs_in'] ['d']
                                            else:
                                                logger.warning("Couldn't get regulates relation and/or regulated term from LD of: {}".format(term))
                                        elif bucket in ["b", "c"]:
                                            regulates_rel, regulated_term = self.get_rel_and_term_in_logical_definitions(term)
                                            if regulates_rel:
                                                # find 'Y subPropertyOf regulates_rel' in RO where Y will be `causally
                                                # upstream of` relation
                                                # Ex. GO:0045944 -> RO:0002213 -> RO:0002304
                                                # edges(RO:0002213) only returns subProperties. Need superProperties
                                                # Gettin super properties
                                                causally_upstream_relation = self.get_causally_upstream_relation(regulates_rel)
                                                # GP-A<-enabled_by-[root MF]-part_of->[regulation of Z]-has_input->GP-B,-causally upstream of (positive/negative effect)->[root MF]-enabled_by->GP-B
                                                ext_target_n = annot_subgraph.add_instance_of_class(ext_target)
                                                anchor_n = annot_subgraph.get_anchor()  # TODO: Gotta find MF. MF no longer anchor if primary term is BP
                                                annot_subgraph.add_edge(anchor_n, INPUT_RELATIONS["has input"], ext_target_n)
                                                root_mf_b_n = annot_subgraph.add_instance_of_class(upt.molecular_function)
                                                annot_subgraph.add_edge(anchor_n, causally_upstream_relation, root_mf_b_n)
                                                annot_subgraph.add_edge(root_mf_b_n, ro.enabled_by, ext_target_n)
                                                # WB:WBGene00001574 GO:1903363 ['happens_during', 'has_regulation_target'] ['c', 'd']
                                            else:
                                                logger.warning("Couldn't get regulates relation from LD of: {}".format(term))

                    else:
                        logger.debug("BAD: {}".format(ext_str))
                    # For annots w/ extensions, this is where we write subgraph to model
                    annot_subgraph.write_to_model(self, evidences)
        # self.extensions_mapper.go_aspector.write_cache()

        if len(self.contributors) == 0 or self.contributors == set(self.default_contributor):
            contributors = set(self.default_contributor)
        elif self.default_contributor in self.contributors:
            contributors = self.contributors - set(self.default_contributor)
        else:
            contributors = self.contributors
        for c in contributors:
            self.declare_contributor(c)
        for pb in self.provided_bys:
            self.declare_provided_by(pb)

    def translate_primary_annotation(self, annotation: CollapsedAssociation):
        gp_id = str(annotation.subject_id())
        term = str(annotation.object_id())
        annot_subgraph = AnnotationSubgraph()

        # TODO: qualifiers are coming through as relation terms now
        for q_term in annotation.qualifiers:
            if ":" in str(q_term):
                # It's a curie (nice!) but we're talking labels
                # q = self.ro_ontology.label(str(q_term)).replace(" ", "_")
                q = self.ontology.label(str(q_term)).replace(" ", "_")
            else:
                q = q_term
            if q == "enables":
                term_n = annot_subgraph.add_instance_of_class(term, is_anchor=True, negated=annotation.negated)
                enabled_by_n = annot_subgraph.add_instance_of_class(gp_id)
                annot_subgraph.add_edge(term_n, "RO:0002333", enabled_by_n)
            elif q == "involved_in":
                mf_n = annot_subgraph.add_instance_of_class(upt.molecular_function)
                enabled_by_n = annot_subgraph.add_instance_of_class(gp_id)
                # term_n = annot_subgraph.add_instance_of_class(term)
                term_n = annot_subgraph.add_instance_of_class(term, is_anchor=True, negated=annotation.negated)
                annot_subgraph.add_edge(mf_n, "RO:0002333", enabled_by_n)
                annot_subgraph.add_edge(mf_n, "BFO:0000050", term_n)
            elif q in ACTS_UPSTREAM_OF_RELATIONS:
                # Look for existing GP <- enabled_by [root MF] -> causally_upstream_of BP
                causally_relation = self.get_causally_upstream_relation(ACTS_UPSTREAM_OF_RELATIONS[q])
                # mf_n = annot_subgraph.add_instance_of_class(upt.molecular_function, is_anchor=True)
                mf_n = annot_subgraph.add_instance_of_class(upt.molecular_function)
                enabled_by_n = annot_subgraph.add_instance_of_class(gp_id)
                # term_n = annot_subgraph.add_instance_of_class(term)
                term_n = annot_subgraph.add_instance_of_class(term, is_anchor=True, negated=annotation.negated)
                annot_subgraph.add_edge(mf_n, "RO:0002333", enabled_by_n)
                annot_subgraph.add_edge(mf_n, causally_relation, term_n)
            else:
                # TODO: should check that existing axiom/triple isn't connected to anything else; length matches exactly
                enabled_by_n = annot_subgraph.add_instance_of_class(gp_id)
                term_n = annot_subgraph.add_instance_of_class(term, is_anchor=True, negated=annotation.negated)
                # annot_subgraph.add_edge(enabled_by_n, self.relations_dict[q], term_n)
                if q == "part_of" and SHEX_HELPER.shape_from_class(term, self.go_aspector) == "CellularComponent":
                    # Using shex_shape function should exclude AnatomicalEntity from getting located_in
                    q = "located_in"
                if q not in self.relations_dict:
                    relation = self.translate_relation_to_ro(q)
                else:
                    relation = self.relations_dict[q]
                annot_subgraph.add_edge(enabled_by_n, relation, term_n)

        with_froms = annotation.with_from()
        if with_froms:
            for wf in with_froms:
                wf_n = annot_subgraph.add_instance_of_class(wf)
                annot_subgraph.add_edge(annot_subgraph.get_anchor(), "RO:0002233", wf_n)

        return annot_subgraph

    def translate_relation_to_ro(self, relation_label):
        # Also check in GO_REL and use xref to RO
        # for n in self.ro_ontology.nodes():
        #     node_label = self.ro_ontology.label(n)
        for n in self.ontology.nodes():
            node_label = self.ontology.label(n)
            if node_label == relation_label.replace("_", " "):
                return n
        # # Is GOREL in go-lego?
        # for n in self.ontology.nodes():
        #     node_label = self.ontology.label(n)
            if node_label == relation_label:
                gorel_node = self.ontology.node(n)
                # What we want will likely be in xref:
                xrefs = gorel_node['meta'].get('xrefs')
                if xrefs and len(xrefs) > 0:
                    for xref in xrefs:
                        val = xref['val']
                        if val.startswith('RO') or val.startswith('BFO'):
                            # print("{} xref'd to {}".format(n, val))
                            return val
                    fallback_rel = xrefs[0]['val']  # default to the the first xref - usually GOREL
                    self.writer.emit_type(URIRef(expand_uri_wrapper(fallback_rel)), OWL.ObjectProperty)
                    self.writer.emit(URIRef(expand_uri_wrapper(fallback_rel)), RDFS.label, Literal(relation_label))
                    return fallback_rel
                # print(gorel_node)  # No such luck so far getting matches

    def get_restrictions(self, term):
        lds = self.ontology.logical_definitions(term)
        restrictions = []
        for ld in lds:
            for r in ld.restrictions:
                # if r[0] in self.ro_ontology.descendants("RO:0002211", reflexive=True):
                if r[0] in self.ontology.descendants("RO:0002211", reflexive=True):
                    restrictions.append(r)
        return restrictions

    def get_rel_and_term_in_logical_definitions(self, term):
        term_restrictions = self.get_restrictions(term)
        if len(term_restrictions) > 0:
            first_restriction = term_restrictions[0]
            regulates_rel = first_restriction[0]
            regulated_term = first_restriction[1]
            return regulates_rel, regulated_term
        else:
            return None, None

    def get_causally_upstream_relation(self, relation):
        regulates = "RO:0002211"
        causally_upstream_relations = []
        if relation == regulates or regulates in self.ontology.ancestors(relation):
            for p in self.ontology.parents(relation,
                                              relations=['subPropertyOf']):
                # For GO:0045944 this is grabbing both RO:0002304 and RO:0002211
                # Need specifically RO:0002304; how to specify?
                #   regulates_rel will only ever be regulates, positively regulates, or
                #   negatively regulates. If regulates in parents, grab other term
                if not p == regulates:
                    causally_upstream_relations.append(p)
        # input relations could have some unique logical difference (e.g. positively_regulates vs acts_upstream_of)
        else:
            if relation in self.ENABLES_O_RELATION_LOOKUP:
                return self.ENABLES_O_RELATION_LOOKUP[relation]
            else:
                property_chain_axioms = self.ontology.get_property_chain_axioms(relation)
                causally_upstream_relation = property_chain_axioms[0].chain_predicate_ids[1]
                self.ENABLES_O_RELATION_LOOKUP[relation] = causally_upstream_relation
                return causally_upstream_relation
        if len(causally_upstream_relations) > 0:
            return causally_upstream_relations[0]
        else:
            return None


class ReferencePreference:
    def __init__(self):
        # List order in python should be persistent
        self.order_of_prefix_preference = [
            "PMID",
            "GO_REF",
            "DOI"
        ]

    def pick(self, references):
        for pfx in self.order_of_prefix_preference:
            for ref in references:
                if ref.upper().startswith(pfx.upper()):
                    return ref


class CamTurtleRdfWriter(TurtleRdfWriter):
    def __init__(self, modeltitle, store=None, model_id: str=None):
        base = "http://model.geneontology.org"
        if model_id is not None:
            self.base = URIRef(model_id, base=base)
        else:
            self.base = genid(base=base)
        if store is not None:
            graph = rdflib.Graph(identifier=self.base, store=store)
        else:
            graph = rdflib.Graph(identifier=self.base)
        self.graph = graph
        self.graph.bind("owl", OWL)
        self.graph.bind("obo", "http://purl.obolibrary.org/obo/")
        self.graph.bind("dc", DC)
        self.graph.bind("rdfs", RDFS)

        self.graph.add((self.base, RDF.type, OWL.Ontology))

        # Model attributes TODO: Should move outside init
        self.graph.add((self.base, DC.date, Literal(datetime.date.today().isoformat())))
        self.graph.add((self.base, DC.title, Literal(modeltitle)))
        self.graph.add((self.base, URIRef("http://geneontology.org/lego/modelstate"), Literal("development")))
        self.graph.add((self.base, OWL.versionIRI, self.base))


class AnnotonCamRdfTransform(CamRdfTransform):
    def __init__(self, writer=None):
        CamRdfTransform.__init__(self, writer)
        self.annotons = []
        self.classes = []
        self.evidences = []
        self.ev_ids = []
        self.bp_id = None

    # TODO Remove "find" feature
    def find_or_create_evidence_id(self, evidence):
        for existing_evidence in self.evidences:
            if evidence.evidence_code == existing_evidence.evidence_code and set(evidence.references) == set(existing_evidence.references):
                if existing_evidence.id is None:
                    existing_evidence.id = genid(base=self.writer.base + '/')
                    self.ev_ids.append(existing_evidence.id)
                return existing_evidence.id
        return self.create_evidence(evidence)

    def create_evidence(self, evidence):
        # Use/figure out standard for creating URIs
        # Find minerva code to generate URI, add to Noctua doc
        ev_id = genid(base=self.writer.base + '/')
        evidence.id = ev_id
        # ev_cls = self.eco_class(self.uri(evidence.evidence_code))
        # ev_cls = self.eco_class(evidence.evidence_code) # This is already ECO:##### due to a GPAD being used
        ev_cls = self.uri(evidence.evidence_code)
        self.emit_type(ev_id, OWL.NamedIndividual)
        self.emit_type(ev_id, ev_cls)
        self.emit(ev_id, DC.date, Literal(evidence.date))
        if evidence.with_from:
            self.emit(ev_id, URIRef("http://geneontology.org/lego/evidence-with"), Literal(evidence.with_from))
        for c in evidence.contributors:
            self.emit(ev_id, DC.contributor, Literal(c))
        for ref_to_emit in evidence.references:
            o = Literal(ref_to_emit)  # Needs to go into Noctua like 'PMID:####' rather than full URL
            self.emit(ev_id, HAS_SUPPORTING_REFERENCE, o)
        self.evidences.append(evidence)
        return evidence.id

    # Use only for OWLAxioms
    # There are two of these methods. AnnotonCamRdfTransform.find_bnode and GoCamModel.find_bnode. Which one is used?
    def find_bnode(self, triple):
        (subject,predicate,object_id) = triple
        s_triples = self.writer.graph.triples((None, OWL.annotatedSource, subject))
        s_bnodes = [s for s,p,o in s_triples]
        p_triples = self.writer.graph.triples((None, OWL.annotatedProperty, predicate))
        p_bnodes = [s for s,p,o in p_triples]
        o_triples = self.writer.graph.triples((None, OWL.annotatedTarget, object_id))
        o_bnodes = [s for s,p,o in o_triples]
        bnodes = set(s_bnodes) & set(p_bnodes) & set(o_bnodes)
        if len(bnodes) > 0:
            return list(bnodes)[0]

    def emit_axiom(self, source_id, property_id, target_id):
        stmt_id = self.blanknode()
        self.emit_type(stmt_id, OWL.Axiom)
        self.emit(stmt_id, OWL.annotatedSource, source_id)
        self.emit(stmt_id, OWL.annotatedProperty, property_id)
        self.emit(stmt_id, OWL.annotatedTarget, target_id)
        return stmt_id

    def find_annotons(self, enabled_by, annotons_list=None):
        found_annotons = []
        if annotons_list is not None:
            annotons = annotons_list
        else:
            annotons = self.annotons
        for annoton in annotons:
            if annoton.enabled_by == enabled_by:
                found_annotons.append(annoton)
        return found_annotons

    def add_individual(self, individual_id, annoton):
        obj_uri = self.uri(individual_id)
        if individual_id not in annoton.individuals:
            tgt_id = genid(base=self.writer.base + '/')
            annoton.individuals[individual_id] = tgt_id
            self.emit_type(tgt_id, obj_uri)
            self.emit_type(tgt_id, OWL.NamedIndividual)
        else:
            tgt_id = annoton.individuals[individual_id]



from ontobio.io import assocparser
from ontobio.io import gafparser

def test_correct_taxon_mapping():
    parser = gafparser.GafParser()
    mapped = parser._taxon_id("taxon:123", assocparser.SplitLine("", [""] * 17, "taxon:123"))

    assert mapped == "NCBITaxon:123"

def test_invalid_mapping():
    parser = gafparser.GafParser()
    mapped = parser._taxon_id("taxon:", assocparser.SplitLine("", [""] * 17, "taxon:"))

    assert mapped == None

def test_already_normalized_mapping():
    parser = gafparser.GafParser()
    mapped = parser._taxon_id("NCBITaxon:123", assocparser.SplitLine("", [""] * 17, "NCBITaxon:123"))

    assert mapped == "NCBITaxon:123"



from ontobio.io import assocwriter
from ontobio.io import gafparser, gpadparser
from ontobio.model.association import GoAssociation, Curie, Subject, Term, ConjunctiveSet, Evidence, ExtensionUnit, Date, Aspect, Provider
import json
import io


def test_gaf_writer():
    association = GoAssociation(
        source_line="",
        subject=Subject(
            id=Curie("PomBase", "SPAC25B8.17"),
            label="ypf1",
            type=["protein"],
            fullname=["intramembrane aspartyl protease of the perinuclear ER membrane Ypf1 (predicted)"],
            synonyms=["ppp81"],
            taxon=Curie("NCBITaxon", "4896")
        ),
        object=Term(
            id=Curie("GO", "0000006"),
            taxon=Curie("NCBITaxon", "4896")
        ),
        negated=False,
        qualifiers=[],
        aspect=Aspect("C"),
        relation=Curie("BFO", "0000050"),
        interacting_taxon=Curie("NCBITaxon", "555"),
        evidence=Evidence(
            type=Curie("ECO", "0000266"),
            has_supporting_reference=[Curie("GO_REF", "0000024")],
            with_support_from=[ConjunctiveSet(
                elements=[Curie("SGD", "S000001583")]
            )]
        ),
        provided_by=Provider("PomBase"),
        date=Date(year="2015", month="03", day="05", time=""),
        subject_extensions=[
            ExtensionUnit(
                relation=Curie("rdfs", "subClassOf"),
                term=Curie("UniProtKB", "P12345")
            )
        ],
        object_extensions=[
            ConjunctiveSet(elements=[
                ExtensionUnit(
                    relation=Curie("BFO", "0000050"),
                    term=Curie("X", "1")
                )
            ])
        ],
        properties=dict()
    )
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out)
    # `out` will get written with gaf lines from the above assocation object
    expected = "PomBase\tSPAC25B8.17\typf1\t\tGO:0000006\tGO_REF:0000024\tISO\tSGD:S000001583\tC\tintramembrane aspartyl protease of the perinuclear ER membrane Ypf1 (predicted)\tppp81\tprotein\ttaxon:4896|taxon:555\t20150305\tPomBase\tpart_of(X:1)\tUniProtKB:P12345"
    writer.write_assoc(association)
    print(out.getvalue())
    gaf = [line.strip("\n") for line in out.getvalue().split("\n") if not line.startswith("!")][0]
    assert expected == gaf

def test_full_taxon_field_single_taxon():
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out)

    taxon_field = writer._full_taxon_field("taxon:12345", None)
    assert "taxon:12345" == taxon_field

def test_full_taxon_field_interacting():
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out)

    taxon_field = writer._full_taxon_field("taxon:12345", "taxon:6789")
    assert "taxon:12345|taxon:6789" == taxon_field

def test_full_taxon_empty_string_interacting_taxon():
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out)

    taxon_field = writer._full_taxon_field("taxon:12345", "")
    assert "taxon:12345" == taxon_field

def test_negated_qualifers():
    gaf = ["PomBase", "SPBC11B10.09", "cdc2", "NOT", "GO:0007275", "PMID:21873635", "ISO", "PANTHER:PTN000623979|TAIR:locus:2099478", "P", "Cyclin-dependent kinase 1", "UniProtKB:P04551|PTN000624043", "protein", "taxon:284812", "20170228", "GO_Central", "", ""]
    parser = gafparser.GafParser()
    result = parser.parse_line("\t".join(gaf))
    writer = assocwriter.GafWriter()
    parsed = writer.as_tsv(result.associations[0])
    print(parsed)
    assert parsed[3] == "NOT"

    writer = assocwriter.GpadWriter()
    parsed = writer.as_tsv(result.associations[0])
    print(parsed)
    assert parsed[2] == "NOT|involved_in"

def test_roundtrip():
    """
    Start with a line, parse it, then write it. The beginning line should be the same as what was written.
    """
    line = "PomBase\tSPAC25B8.17\typf1\t\tGO:0000006\tGO_REF:0000024\tISO\tSGD:S000001583\tC\tintramembrane aspartyl protease of the perinuclear ER membrane Ypf1 (predicted)\tppp81\tprotein\ttaxon:999|taxon:888\t20150305\tPomBase\tpart_of(X:1)\tUniProtKB:P12345"
    parser = gafparser.GafParser()
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out)
    assoc = parser.parse_line(line).associations[0]
    writer.write_assoc(assoc)
    gaf = [line for line in out.getvalue().split("\n") if not line.startswith("!")][0]
    assert line == gaf

    # Single taxon
    line = "PomBase\tSPAC25B8.17\typf1\t\tGO:0000006\tGO_REF:0000024\tISO\tSGD:S000001583\tC\tintramembrane aspartyl protease of the perinuclear ER membrane Ypf1 (predicted)\tppp81\tprotein\ttaxon:1111\t20150305\tPomBase\tpart_of(X:1)\tUniProtKB:P12345"
    parser = gafparser.GafParser()
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out)
    assoc = parser.parse_line(line).associations[0]
    writer.write_assoc(assoc)
    gaf = [line for line in out.getvalue().split("\n") if not line.startswith("!")][0]
    assert line == gaf

def test_gpad_qualifier_removed_in_gaf_2_1():
    # Qualifier is `part_of` and should be returned blank instead of removing the whole line
    line = "PomBase\tSPBC1348.01\tpart_of\tGO:0009897\tGO_REF:0000051\tECO:0000201\t\t\t20060201\tPomBase\t\t"
    parser = gpadparser.GpadParser()
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out, version="2.1")  # Write out to gaf 2.1

    assoc = parser.parse_line(line).associations[0]
    writer.write_assoc(assoc)
    gpad_to_gaf_line = [line for line in out.getvalue().split("\n") if not line.startswith("!")][0]
    assert gpad_to_gaf_line.split("\t")[3] == ""

    # Test with a `NOT`
    line = "PomBase\tSPBC1348.01\tNOT|part_of\tGO:0009897\tGO_REF:0000051\tECO:0000201\t\t\t20060201\tPomBase\t\t"
    parser = gpadparser.GpadParser()
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out, version="2.1")  # Write out to gaf 2.1

    assoc = parser.parse_line(line).associations[0]
    writer.write_assoc(assoc)
    gpad_to_gaf_line = [line for line in out.getvalue().split("\n") if not line.startswith("!")][0]
    assert gpad_to_gaf_line.split("\t")[3] == "NOT"

def test_gaf2_2_qualifier_to_gaf2_1():
    # Qualifier is `part_of` and should be returned blank instead of removing the whole line
    line = "WB\tWBGene00000001\taap-1\tinvolved_in\tGO:0008286\tWB_REF:WBPaper00005614|PMID:12393910\tIMP\t\tP\t\tY110A7A.10\tgene\ttaxon:6239\t20060302\tWB\t\t"
    parser = gafparser.GafParser()
    parser.version = "2.2"
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out, version="2.1")  # Write out to gaf 2.1

    assoc = parser.parse_line(line).associations[0]
    writer.write_assoc(assoc)
    gpad_to_gaf_line = [line for line in out.getvalue().split("\n") if not line.startswith("!")][0]
    assert gpad_to_gaf_line.split("\t")[3] == ""

    # Test with a `NOT`
    line = "WB\tWBGene00000001\taap-1\tNOT|involved_in\tGO:0008286\tWB_REF:WBPaper00005614|PMID:12393910\tIMP\t\tP\t\tY110A7A.10\tgene\ttaxon:6239\t20060302\tWB\t\t"
    parser = gafparser.GafParser()
    parser.version = "2.2"
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out, version="2.1")  # Write out to gaf 2.1

    assoc = parser.parse_line(line).associations[0]
    writer.write_assoc(assoc)
    gpad_to_gaf_line = [line for line in out.getvalue().split("\n") if not line.startswith("!")][0]
    assert gpad_to_gaf_line.split("\t")[3] == "NOT"

def test_writing_to_gaf_2_2():
    line = "WB\tWBGene00000001\taap-1\tinvolved_in\tGO:0008286\tWB_REF:WBPaper00005614|PMID:12393910\tIMP\t\tP\t\tY110A7A.10\tgene\ttaxon:6239\t20060302\tWB\t\t"
    parser = gafparser.GafParser()
    parser.version = "2.2"
    assoc = parser.parse_line(line).associations[0] # type: GoAssociation

    gaf_22_out = assoc.to_gaf_2_2_tsv()
    assert gaf_22_out[3] == "involved_in"

    # With NOT
    line = "WB\tWBGene00000001\taap-1\tNOT|involved_in\tGO:0008286\tWB_REF:WBPaper00005614|PMID:12393910\tIMP\t\tP\t\tY110A7A.10\tgene\ttaxon:6239\t20060302\tWB\t\t"
    parser = gafparser.GafParser()
    parser.version = "2.2"

    assoc = parser.parse_line(line).associations[0] # type: GoAssociation

    gaf_22_out = assoc.to_gaf_2_2_tsv()
    assert gaf_22_out[3] == "NOT|involved_in"

def test_gaf_2_2_extensions():
    line = "WB\tWBGene00000001\taap-1\tinvolved_in\tGO:0008286\tWB_REF:WBPaper00005614|PMID:12393910\tIMP\t\tP\t\tY110A7A.10\tgene\ttaxon:6239\t20060302\tWB\tpart_of(EMAPA:17972),part_of(CL:0000018)\t"
    parser = gafparser.GafParser()
    parser.version = "2.2"
    assoc = parser.parse_line(line).associations[0]

    gaf_22_out = assoc.to_gaf_2_2_tsv()
    assert gaf_22_out[15] == "part_of(EMAPA:17972),part_of(CL:0000018)"


def test_full_gaf_2_2_write():
    line = "WB\tWBGene00000001\taap-1\tinvolved_in\tGO:0008286\tWB_REF:WBPaper00005614|PMID:12393910\tIMP\t\tP\t\tY110A7A.10\tgene\ttaxon:6239\t20060302\tWB\t\t"
    parser = gafparser.GafParser()
    parser.version = "2.2"
    out = io.StringIO()
    writer = assocwriter.GafWriter(file=out, version="2.2")

    assoc = parser.parse_line(line).associations[0]
    writer.write_assoc(assoc)
    out_line = [line for line in out.getvalue().split("\n") if not line.startswith("!")][0]
    assert out_line.split("\t") == line.split("\t")

def test_gaf_to_gpad2():
    line = "PomBase\tSPAC25B8.17\typf1\t\tGO:0000006\tGO_REF:0000024\tISO\tSGD:S000001583\tC\tintramembrane aspartyl protease of the perinuclear ER membrane Ypf1 (predicted)\tppp81\tprotein\ttaxon:999|taxon:888\t20150305\tPomBase\tpart_of(X:1)\tUniProtKB:P12345"
    parser = gafparser.GafParser()
    out = io.StringIO()
    writer = assocwriter.GpadWriter(version=assocwriter.GPAD_2_0, file=out)

    assoc = parser.parse_line(line).associations[0]
    writer.write_assoc(assoc)

    lines = out.getvalue().split("\n")
    assert lines[0] == "!gpa-version: 2.0"
    assert lines[1] == "PomBase:SPAC25B8.17\t\tBFO:0000050\tGO:0000006\tGO_REF:0000024\tECO:0000266\tSGD:S000001583\tNCBITaxon:888\t2015-03-05\tPomBase\tBFO:0000050(X:1)\t"

    line = "PomBase\tSPAC25B8.17\typf1\tNOT\tGO:0000006\tGO_REF:0000024\tISO\tSGD:S000001583\tC\tintramembrane aspartyl protease of the perinuclear ER membrane Ypf1 (predicted)\tppp81\tprotein\ttaxon:999|taxon:888\t20150305\tPomBase\tpart_of(X:1)\tUniProtKB:P12345"
    parser = gafparser.GafParser()
    out = io.StringIO()
    writer = assocwriter.GpadWriter(version=assocwriter.GPAD_2_0, file=out)

    assoc = parser.parse_line(line).associations[0]
    writer.write_assoc(assoc)

    lines = out.getvalue().split("\n")
    assert lines[0] == "!gpa-version: 2.0"
    assert lines[1] == "PomBase:SPAC25B8.17\tNOT\tBFO:0000050\tGO:0000006\tGO_REF:0000024\tECO:0000266\tSGD:S000001583\tNCBITaxon:888\t2015-03-05\tPomBase\tBFO:0000050(X:1)\t"

def test_writing_assoc_properties():
    line = "MGI:MGI:1922721\t\tRO:0002327\tGO:0019904\tMGI:MGI:3769586|PMID:17984326\tECO:0000353\tPR:Q0KK55\t\t2010-12-01\tMGI\tBFO:0000066(EMAPA:17787),RO:0002233(MGI:MGI:1923734)\tcreation-date=2008-02-07|modification-date=2010-12-01|comment=v-KIND domain binding of Kndc1;MGI:1923734|contributor-id=http://orcid.org/0000-0003-2689-5511|contributor-id=http://orcid.org/0000-0003-3394-9805"
    parser = gpadparser.GpadParser()
    parser.version = "2.0"
    out = io.StringIO()
    writer = assocwriter.GpadWriter(file=out, version="2.0")  # Write back out to gpad 2.0

    assoc = parser.parse_line(line).associations[0]
    writer.write_assoc(assoc)
    written_gpad_line = [line for line in out.getvalue().split("\n") if not line.startswith("!")][0]
    written_props = written_gpad_line.split("\t")[11]
    assert len(written_props.split("|")) == 5



import re
import enum
import json


from dataclasses import dataclass
from typing import NewType, List, Dict
from prefixcommons import curie_util

from ontobio.model import association
from ontobio.rdfgen import relations

import functools
import logging
import copy

prefix_context = {key: value for context in curie_util.default_curie_maps + [curie_util.read_biocontext("go_context")] for key, value in context.items()}

aspect_relation_map = {
    "F": "http://purl.obolibrary.org/obo/RO_0002327",
    "P": "http://purl.obolibrary.org/obo/RO_0002331",
    "C": "http://purl.obolibrary.org/obo/BFO_0000050"
}

relation_aspect_map = {
    "http://purl.obolibrary.org/obo/RO_0002327": "F",
    "http://purl.obolibrary.org/obo/RO_0002326": "F",
    "http://purl.obolibrary.org/obo/RO_0002264": "P",
    "http://purl.obolibrary.org/obo/RO_0004032": "P",
    "http://purl.obolibrary.org/obo/RO_0004033": "P",
    "http://purl.obolibrary.org/obo/RO_0002263": "P",
    "http://purl.obolibrary.org/obo/RO_0004034": "P",
    "http://purl.obolibrary.org/obo/RO_0004035": "P",
    "http://purl.obolibrary.org/obo/RO_0002331": "P",
    "http://purl.obolibrary.org/obo/BFO_0000050": "C",
    "http://purl.obolibrary.org/obo/RO_0002432": "C",
    "http://purl.obolibrary.org/obo/RO_0002325": "C"
}

Uri = NewType("Uri", str)
Gaf = NewType("Gaf", List)

# This goes like:
# Turn gaferencer json into Dict[AnnotationKey, InferenceValue]
# For a gaf line: gaf --> List[AnnotationKey]
# Use gaferencer dict x gaf annotation key --> List[InferenceValue]
# gaf x gaferencer --> List[Inferences] --> List[gaf]

@dataclass(unsafe_hash=True)
class RelationTo:
    relation: Uri
    term: Uri

@dataclass(unsafe_hash=True)
class AnnotationKey:
    relation_to: RelationTo
    taxon: Uri
    extension: association.ConjunctiveSet

@dataclass(unsafe_hash=True)
class InferenceValue:
    satisfiable: bool
    taxon_problem: bool
    inferences: List[RelationTo]

ProblemType = enum.Enum("ProblemType", {"TAXON": "taxon", "EXTENSION": "extension"})


@dataclass
class InferenceResult:
    inferred_gafs: List[Gaf]
    problem: ProblemType


def load_gaferencer_inferences_from_file(gaferencer_out) -> Dict[AnnotationKey, InferenceValue]:

    gaferencer_out_dict = dict()
    try:
        logging.warning("gaferencer_out = {}".format(gaferencer_out))
        with open(gaferencer_out) as gaferencer_file:
            gaferencer_out_dict = json.load(gaferencer_file)
    except Exception as e:
        logging.warning("Could not load file {}: {}".format(gaferencer_out, str(e)))
        return None

    return build_annotation_inferences(gaferencer_out_dict)

def load_gaferencer_inferences_from_files(gaferencer_out_list) -> Dict[AnnotationKey, InferenceValue]:
    return functools.reduce(load_gaferencer_inferences_from_file, gaferencer_out_list, dict())


def build_annotation_inferences(gaferencer_out: List[Dict]) -> Dict[AnnotationKey, InferenceValue]:
    inferences = dict()  # type Dict[AnnotationKey, InferenceValue]
    for gaference in gaferencer_out:
        key = AnnotationKey(
            RelationTo(
                gaference["annotation"]["annotation"]["relation"],
                gaference["annotation"]["annotation"]["term"]),
            gaference["annotation"]["taxon"],
            make_conjunctions(gaference["annotation"]["extension"])
        )
        value = InferenceValue(
            gaference["satisfiable"],
            gaference["taxonProblem"],
            [RelationTo(inf["relation"], inf["term"]) for inf in gaference["inferences"]]
        )
        inferences[key] = value
    return inferences

def produce_inferences(gaf: association.GoAssociation, inference_table: Dict[AnnotationKey, InferenceValue]) -> List:
    keys = make_keys_from_gaf(gaf)  # type: List[AnnotationKey]
    results = []  # type: List[InferenceResult]
    for key in keys:
        inferred_value = inference_table.get(key, None)
        if inferred_value != None:
            inferred_gafs_result = gaf_inferences_from_value(gaf, inferred_value)  # type: InferenceResult
            results.append(inferred_gafs_result)

    return results

def lookup_relation(label):
    label = label.replace('_', ' ')

    # Return the cached label -> URI or None
    if label in relations.label_relation_lookup():
        return relations.label_relation_lookup()[label]
    else:
        return None

def make_conjunctions(extension: List) -> association.ConjunctiveSet:
    extension_units = []  # type: List[association.ExtensionUnit]
    for unit in extension:
        extension_units.append(association.ExtensionUnit(unit["relation"], unit["term"]))

    return association.ConjunctiveSet(frozenset(extension_units))

relation_tuple = re.compile(r"(.+)\((.+)\)")
def make_keys_from_gaf(gaf: association.GoAssociation) -> List[AnnotationKey]:

    term = curie_util.expand_uri(str(gaf.object.id), cmaps=[prefix_context])
    relation = curie_util.expand_uri(str(gaf.relation), cmaps=[prefix_context])
    taxon = curie_util.expand_uri(str(gaf.object.taxon), cmaps=[prefix_context])
    extensions = gaf.object_extensions # type: List[association.ConjunctiveSet]

    annotation_keys = []  # type: List[AnnotationKey]

    if extensions:
        for conjunction in extensions:
            # Each conjunction is a ConjunctiveSet
            # conjunction is foo(bar),hello(world)
            # Create a new ConjunctiveSet using a frozenset of the elements instead of a list
            frozen_conjunction = association.ConjunctiveSet(frozenset(conjunction.elements))
            # Build the Key now
            annotation_keys.append(AnnotationKey(RelationTo(relation, term), taxon, frozen_conjunction))
    else:
        annotation_keys.append(AnnotationKey(RelationTo(relation, term), taxon, association.ConjunctiveSet(frozenset([]))))

    return annotation_keys


def gaf_inferences_from_value(original_gaf: association.GoAssociation, inferred_value: InferenceValue) -> InferenceResult:
    # This can be successful, where we materialize more than one inference gafs
    # Or it can go wrong, where it's not satisfiable via taxon problems or wrong extensions
    if not inferred_value.satisfiable:
        # Then it's bad
        problem = ProblemType.TAXON if inferred_value.taxon_problem else ProblemType.EXTENSION
        return InferenceResult([], problem)

    # At this point it has inferences
    inferred_gafs = []  # type: List[association.GoAssociation]
    for inference in inferred_value.inferences:
        new_gaf = copy.deepcopy(original_gaf)
        goterm = inference.term.rsplit("/", maxsplit=1)[1].replace("_", ":")
        aspect = relation_aspect_map[inference.relation]
        new_gaf.object.id = goterm
        new_gaf.aspect = aspect
        new_gaf.object_extensions = []
        inferred_gafs.append(new_gaf)

    return InferenceResult(inferred_gafs, None)



from ontobio.io.gpadparser import GpadParser, to_association
from ontobio.io import assocparser
from ontobio.model import association
from ontobio.model.association import ConjunctiveSet, ExtensionUnit, Curie

import yaml

POMBASE = "tests/resources/truncated-pombase.gpad"

def test_skim():
    p = GpadParser()
    results = p.skim(open(POMBASE,"r"))
    print(str(results))


def test_parse():
    p = GpadParser(config=assocparser.AssocParserConfig(group_metadata=yaml.load(open("tests/resources/mgi.dataset.yaml"), Loader=yaml.FullLoader)))
    test_gpad_file = "tests/resources/mgi.test.gpad"
    results = p.parse(open(test_gpad_file, "r"))
    print(p.report.to_markdown())


def test_parse_1_2():
    report = assocparser.Report(group="unknown", dataset="unknown")
    vals = [
        "MGI",
        "MGI:1918911",
        "enables",
        "GO:0003674",
        "MGI:MGI:2156816|GO_REF:0000015",
        "ECO:0000307",
        "",
        "",
        "20100209",
        "MGI",
        "",
        "creation-date=2020-09-17|modification-date=2020-09-17|contributor-id=http://orcid.org/0000-0003-2689-5511"
    ]
    result = to_association(list(vals), report=report, version="1.2")
    assert result.skipped == 0
    assert len([m for m in result.report.messages if m["level"] == "ERROR"]) == 0
    assert len(result.associations) == 1

def test_parse_interacting_taxon():
    report = assocparser.Report(group="unknown", dataset="unknown")
    vals = [
        "MGI",
        "MGI:1918911",
        "enables",
        "GO:0003674",
        "MGI:MGI:2156816|GO_REF:0000015",
        "ECO:0000307",
        "",
        "taxon:5678",
        "20100209",
        "MGI",
        "",
        "creation-date=2020-09-17|modification-date=2020-09-17|contributor-id=http://orcid.org/0000-0003-2689-5511"
    ]
    result = to_association(list(vals), report=report, version="1.2")
    assert result.associations[0].interacting_taxon == Curie(namespace="NCBITaxon", identity="5678")

def test_duplicate_key_annot_properties():
    properties_str = "creation-date=2008-02-07|modification-date=2010-12-01|comment=v-KIND domain binding of Kndc1;MGI:1923734|contributor-id=http://orcid.org/0000-0003-2689-5511|contributor-id=http://orcid.org/0000-0003-3394-9805"
    prop_list = association.parse_annotation_properties(properties_str)
    contributor_ids = [value for key, value in prop_list if key == "contributor-id"]
    assert set(contributor_ids) == {"http://orcid.org/0000-0003-2689-5511", "http://orcid.org/0000-0003-3394-9805"}

def test_parse_2_0():
    version = "2.0"
    report = assocparser.Report(group="unknown", dataset="unknown")
    vals = [
        "MGI:MGI:1918911",
        "",
        "RO:0002327",
        "GO:0003674",
        "MGI:MGI:2156816|GO_REF:0000015",
        "ECO:0000307",
        "",
        "",
        "2020-09-17",
        "MGI",
        "",
        "creation-date=2020-09-17|modification-date=2020-09-17|contributor-id=http://orcid.org/0000-0003-2689-5511"
    ]
    result = to_association(list(vals), report=report, version=version)
    assert result.skipped == 0
    assert len([m for m in result.report.messages if m["level"] == "ERROR"]) == 0
    assert len(result.associations) == 1

    # Annotation_Extensions
    vals[10] = "BFO:0000066(CL:0000010),GOREL:0001004(CL:0000010)"
    result = to_association(list(vals), report=report, version=version)
    assert result.associations[0].object_extensions == [ConjunctiveSet([
            ExtensionUnit(Curie("BFO", "0000066"), Curie("CL", "0000010")),
            ExtensionUnit(Curie("GOREL", "0001004"), Curie("CL", "0000010"))
        ])]

    # With_or_From
    vals[6] = "PR:Q505B8|PR:Q8CHK4"
    result = to_association(list(vals), report=report, version=version)
    assert result.associations[0].evidence.with_support_from == [
        ConjunctiveSet([Curie("PR", "Q505B8")]),
        ConjunctiveSet([Curie("PR", "Q8CHK4")])
    ]

    # Interacting taxon - this example should fail
    vals[7] = "Staphylococcus aureus ; NCBITaxon:1280"
    result = to_association(list(vals), report=report, version=version)
    assert len([m for m in result.report.messages if m["level"] == "ERROR"]) > 0
    assert len(result.associations) == 0
    # Now test valid interacting taxon value
    vals[7] = "NCBITaxon:1280"
    result = to_association(list(vals), report=report, version=version)
    assert result.associations[0].interacting_taxon == Curie("NCBITaxon", "1280")

    # Confirm non-"MGI:MGI:" IDs will parse
    vals[0] = "WB:WBGene00001189"
    result = to_association(list(vals), report=report, version=version)
    assert result.associations[0].subject.id == Curie("WB", "WBGene00001189")

    # Test annotation property retrieval
    contributors = result.associations[0].annotation_property_values(property_key="contributor-id")
    assert set(contributors) == {"http://orcid.org/0000-0003-2689-5511"}
